#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.update.py
"""Functions to update various columns in the SQL database"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics


import numpy as np
from os.path import join as join_path
import pydicom as dicom
from dvha.db.sql_connector import DVH_SQL
from dvha.tools import roi_geometry as roi_geom
from dvha.tools import roi_formatter as roi_form
from dvha.tools.errors import push_to_log
from mlca.mlc_analyzer import Beam as BeamAnalyzer
from dvha.tools.utilities import calc_stats, sample_roi


def centroid(study_instance_uid, roi_name):
    """Recalculate the centroid of an roi based on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi
    """

    coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )

    roi = roi_form.get_planes_from_string(coordinates_string[0][0])
    data = roi_geom.centroid(roi)

    data = [str(round(v, 3)) for v in data]

    update_dvhs_table(study_instance_uid, roi_name, "centroid", ",".join(data))


def cross_section(study_instance_uid, roi_name):
    """Recalculate the centroid of an roi based on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi

    """

    coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )

    roi = roi_form.get_planes_from_string(coordinates_string[0][0])
    area = roi_geom.cross_section(roi)

    for key in ["max", "median"]:
        update_dvhs_table(
            study_instance_uid, roi_name, "cross_section_%s" % key, area[key]
        )


def spread(study_instance_uid, roi_name):
    """Recalculate the spread of an roi based on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi

    """

    coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )

    roi = roi_form.get_planes_from_string(coordinates_string[0][0])
    data = roi_geom.spread(roi)

    data = [str(round(v / 10.0, 3)) for v in data]

    for i, column in enumerate(["spread_x", "spread_y", "spread_z"]):
        update_dvhs_table(study_instance_uid, roi_name, column, data[i])


def dist_to_ptv_centroids(study_instance_uid, roi_name, pre_calc=None):
    """Recalculate the OAR-to-PTV centroid distance based on data in the
    SQL DB. Optionally provide pre-calculated centroid of combined PTV

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi
    pre_calc : np.ndarray
        Return from get_treatment_volume_centroid

    """

    oar_centroid_string = query(
        "dvhs",
        "centroid",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )
    oar_centroid = np.array(
        [float(i) for i in oar_centroid_string[0][0].split(",")]
    )

    ptv_centroid = pre_calc
    if ptv_centroid is None:
        tv = get_total_treatment_volume_of_study(study_instance_uid)
        ptv_centroid = get_treatment_volume_centroid(tv)
    data = float(np.linalg.norm(ptv_centroid - oar_centroid)) / 10.0

    update_dvhs_table(
        study_instance_uid,
        roi_name,
        "dist_to_ptv_centroids",
        round(float(data), 3),
    )


def min_distances(study_instance_uid, roi_name, pre_calc=None):
    """Recalculate the min, mean, median, and max PTV distances an roi based
    on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi
    pre_calc : list, optional
        coordinates of combined PTV, return from get_treatment_volume_coord

    """

    oar_coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )

    treatment_volume_roi = pre_calc
    treatment_volume_coord = get_treatment_volume_coord(pre_calc)
    if treatment_volume_coord is None:
        with DVH_SQL() as cnx:
            ptv_coordinates_strings = cnx.query(
                "dvhs",
                "roi_coord_string",
                "study_instance_uid = '%s' and roi_type like 'PTV%%'"
                % study_instance_uid,
            )

        ptvs = [
            roi_form.get_planes_from_string(ptv[0])
            for ptv in ptv_coordinates_strings
        ]
        treatment_volume_roi = roi_geom.union(ptvs)
        treatment_volume_coord = roi_form.get_roi_coordinates_from_planes(
            treatment_volume_roi
        )

    oar_coordinates = roi_form.get_roi_coordinates_from_string(
        oar_coordinates_string[0][0]
    )

    treatment_volume_coord = sample_roi(treatment_volume_coord)
    oar_coordinates = sample_roi(oar_coordinates)

    try:
        is_inside = [
            [1, -1][roi_geom.is_point_inside_roi(point, treatment_volume_roi)]
            for point in oar_coordinates
        ]
        data = roi_geom.min_distances_to_target(
            oar_coordinates, treatment_volume_coord, factors=is_inside
        )
    except Exception:
        try:
            treatment_volume_coord = sample_roi(
                treatment_volume_coord, max_point_count=3000
            )
            oar_coordinates = sample_roi(oar_coordinates, max_point_count=3000)
            is_inside = [
                [1, -1][
                    roi_geom.is_point_inside_roi(point, treatment_volume_roi)
                ]
                for point in oar_coordinates
            ]
            data = roi_geom.min_distances_to_target(
                oar_coordinates, treatment_volume_coord, factors=is_inside
            )
        except Exception as e:
            msg = (
                "db.update.min_distances: Error reported for %s with "
                "study_instance_uid %s\n"
                "Skipping PTV distance and DTH calculations for this ROI."
                % (roi_name, study_instance_uid)
            )
            push_to_log(e, msg=msg)
            data = None

    if data is not None:
        try:
            dth = roi_geom.dth(data)
            dth_string = ",".join(["%0.6f" % num for num in dth])

            data_map = {
                "dist_to_ptv_min": round(float(np.min(data)), 2),
                "dist_to_ptv_mean": round(float(np.mean(data)), 2),
                "dist_to_ptv_median": round(float(np.median(data)), 2),
                "dist_to_ptv_max": round(float(np.max(data)), 2),
                "dth_string": dth_string,
            }
        except MemoryError as e:
            msg = (
                "Error reported for %s with study_instance_uid %s\n"
                "Skipping PTV distance and DTH calculations for this ROI."
                % (roi_name, study_instance_uid)
            )
            push_to_log(e, msg=msg)
            data_map = None

        if data_map:
            for key, value in data_map.items():
                update_dvhs_table(study_instance_uid, roi_name, key, value)


def treatment_volume_overlap(study_instance_uid, roi_name, pre_calc=None):
    """Recalculate the PTV overlap of an roi based on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi
    pre_calc : dict, optional
        union of PTVs, return from get_total_treatment_volume_of_study

    """

    oar_coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )
    oar = roi_form.get_planes_from_string(oar_coordinates_string[0][0])

    treatment_volume = pre_calc
    if treatment_volume is None:
        treatment_volume = get_total_treatment_volume_of_study(
            study_instance_uid
        )

    overlap = roi_geom.overlap_volume(oar, treatment_volume)
    update_dvhs_table(
        study_instance_uid, roi_name, "ptv_overlap", round(float(overlap), 2)
    )


def volumes(study_instance_uid, roi_name):
    """Recalculate the volume of an roi based on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi

    """

    coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )

    roi = roi_form.get_planes_from_string(coordinates_string[0][0])

    data = roi_geom.volume(roi)

    update_dvhs_table(
        study_instance_uid, roi_name, "volume", round(float(data), 2)
    )


def surface_area(study_instance_uid, roi_name):
    """Recalculate the surface area of an roi based on data in the SQL DB.

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    roi_name : str
        name of roi

    """

    coordinates_string = query(
        "dvhs",
        "roi_coord_string",
        "study_instance_uid = '%s' and roi_name = '%s'"
        % (study_instance_uid, roi_name),
    )

    roi = roi_form.get_planes_from_string(coordinates_string[0][0])

    data = roi_geom.surface_area(roi, coord_type="sets_of_points")

    update_dvhs_table(
        study_instance_uid, roi_name, "surface_area", round(float(data), 2)
    )


def update_dvhs_table(study_instance_uid, roi_name, column, value):
    """Generic function to update a value in the DVHs table

    Parameters
    ----------
    study_instance_uid : str
        study instance uid in the SQL table
    roi_name : str
        the roi name associated with the value to be updated
    column : str
        the SQL column of the value to be updated
    value : str, int, float, datetime
        the value to be set, it's type should match the type as specified in
        the SQL table

    """
    with DVH_SQL() as cnx:
        cnx.update(
            "dvhs",
            column,
            value,
            "study_instance_uid = '%s' and roi_name = '%s'"
            % (study_instance_uid, roi_name),
        )


def update_plan_toxicity_grades(cnx, study_instance_uid):
    """Query the toxicities in the DVHs table and update the values in the
    associated plan row(s)

    Parameters
    ----------
    cnx : DVH_SQL
        connection to DVHA SQL database
    study_instance_uid : str
        study_instance_uid in SQL database

    """
    toxicities = cnx.get_unique_values(
        "DVHs",
        "toxicity_grade",
        "study_instance_uid = '%s'" % study_instance_uid,
    )
    toxicities = [t for t in toxicities if t.isdigit()]
    toxicities_str = ",".join(toxicities)
    cnx.update(
        "Plans",
        "toxicity_grades",
        toxicities_str,
        "study_instance_uid = '%s'" % study_instance_uid,
    )


def plan_complexity(cnx, study_instance_uid):
    """

    Parameters
    ----------
    cnx : DVH_SQL
        connection to DVHA SQL database
    study_instance_uid : str
        study_instance_uid in SQL database

    """
    condition = "study_instance_uid = '%s'" % study_instance_uid
    beam_data = query("Beams", "complexity, beam_mu", condition)
    scores = [row[0] for row in beam_data]
    include = [i for i, score in enumerate(scores) if score]
    scores = [score for i, score in enumerate(scores) if i in include]
    beam_mu = [row[1] for i, row in enumerate(beam_data) if i in include]
    plan_mu = np.sum(beam_mu)
    if plan_mu:
        complexity = np.sum(np.multiply(scores, beam_mu)) / plan_mu
        cnx.update(
            "Plans",
            "complexity",
            complexity,
            "study_instance_uid = '%s'" % study_instance_uid,
        )
    else:
        msg = (
            "db.update.plan_complexity: Zero plan MU detected for uid %s"
            % study_instance_uid
        )
        push_to_log(msg=msg)


def beam_complexity(cnx, study_instance_uid):
    """

    Parameters
    ----------
    cnx : DVH_SQL
        connection to DVHA SQL database
    study_instance_uid : str
        study_instance_uid in SQL database

    """

    rt_plan_query = cnx.query(
        "DICOM_Files",
        "folder_path, plan_file",
        "study_instance_uid = '%s'" % study_instance_uid,
    )[0]
    rt_plan_file_path = join_path(rt_plan_query[0], rt_plan_query[1])

    rt_plan = dicom.read_file(rt_plan_file_path)

    column_vars = {
        "area": "area",
        "x_perim": "x_perim",
        "y_perim": "y_perim",
        "complexity": "cmp_score",
        "cp_mu": "cp_mu",
    }
    stat_map = {"min": 5, "mean": 3, "median": 2, "max": 0}

    unit_factor = {
        "area": 100.0,
        "x_perim": 10.0,
        "y_perim": 10.0,
        "complexity": 1.0,
        "cp_mu": 1.0,
    }

    for beam_num, beam in enumerate(rt_plan.BeamSequence):
        try:
            condition = "study_instance_uid = '%s' and beam_number = '%s'" % (
                study_instance_uid,
                (beam_num + 1),
            )
            meterset = float(cnx.query("Beams", "beam_mu", condition)[0][0])
            mlca_data = BeamAnalyzer(beam, meterset, ignore_zero_mu_cp=True)
            mlc_keys = ["area", "x_perim", "y_perim", "cmp_score", "cp_mu"]
            summary_stats = {
                key: calc_stats(mlca_data.summary[key]) for key in mlc_keys
            }

            for c in list(column_vars):
                for s in list(stat_map):
                    value = (
                        summary_stats[column_vars[c]][stat_map[s]]
                        / unit_factor[c]
                    )
                    column = "%s_%s" % (c, s)
                    cnx.update("Beams", column, value, condition)
            cnx.update(
                "Beams",
                "complexity",
                np.sum(mlca_data.summary["cmp_score"]),
                condition,
            )
        except Exception as e:
            msg = (
                "db.update.beam_complexity: MLC Analyzer fail for beam number "
                "%s and uid %s" % ((beam_num + 1), study_instance_uid)
            )
            push_to_log(e, msg=msg)


def update_all_generic(table, func, condition):
    """Generic function to call ``func`` that accepts a DVH_SQL object and
    study_instance_uid. Intended for beam_complexities, plan_complexities,
    update_all_plan_toxicity_grades, etc.

    Parameters
    ----------
    table : str
        SQL table name
    func : callable
        the function to be called with parameters of DVH_SQL and
        study_instance_uid
    condition : str
        optional SQL condition to apply to the study_instance_uid list
        retrieval

    """
    if condition:
        condition = condition[0]
    with DVH_SQL() as cnx:
        uids = cnx.get_unique_values(
            table, "study_instance_uid", condition, return_empty=True
        )
        for uid in uids:
            func(cnx, uid)


def update_all_plan_toxicity_grades(condition=None):
    """Call update_plan_toxicity_grades on all Plan rows

    Parameters
    ----------
    condition : str
        SQL condition

    """
    update_all_generic("Plans", update_plan_toxicity_grades, condition)


def plan_complexities(condition=None):
    """Call plan_complexity on all Plan rows

    Parameters
    ----------
    condition : str
        SQL condition

    """
    update_all_generic("Plans", plan_complexity, condition)


def beam_complexities(condition=None):
    """Call beam_complexity on all Beams rows

    Parameters
    ----------
    condition : str
        SQL condition

    """
    update_all_generic("Beams", beam_complexity, condition)


def update_ptv_data(tv, study_instance_uid):
    """Update ptv related columns based on a total treatment volume

    Parameters
    ----------
    tv : dict
        treatment volume formatted as a "sets of points" object specified in
        tools.roi_geometry
    study_instance_uid : str
        study_instance_uid in SQL database

    """
    ptv_cross_section = roi_geom.cross_section(tv)
    ptv_spread = roi_geom.spread(tv)

    condition = (
        "study_instance_uid = '%s' and roi_type like 'PTV%%'"
        % study_instance_uid
    )
    with DVH_SQL() as cnx:
        max_dose = cnx.get_max_value("dvhs", "max_dose", condition=condition)
        min_dose = cnx.get_min_value("dvhs", "min_dose", condition=condition)

        ptv_data = {
            "ptv_cross_section_max": ptv_cross_section["max"],
            "ptv_cross_section_median": ptv_cross_section["median"],
            "ptv_max_dose": max_dose,
            "ptv_min_dose": min_dose,
            "ptv_spread_x": ptv_spread[0],
            "ptv_spread_y": ptv_spread[1],
            "ptv_spread_z": ptv_spread[2],
            "ptv_surface_area": roi_geom.surface_area(
                tv, coord_type="sets_of_points"
            ),
            "ptv_volume": roi_geom.volume(tv),
        }

        for key, value in ptv_data.items():
            cnx.update(
                "Plans",
                key,
                value,
                "study_instance_uid = '%s'" % study_instance_uid,
            )


def get_total_treatment_volume_of_study(study_instance_uid, ptvs=None):
    """Calculate combined PTV for the provided study_instance_uid

    Parameters
    ----------
    study_instance_uid : str
        study instance uid
    ptvs : list, optional
        names of ptvs (as stored in roi_name column)

    Returns
    -------
    dict
        a "sets of points" dictionary representing the total treatment volume

    """

    condition = (
        "study_instance_uid = '%s' and roi_type like 'PTV%%'"
        % study_instance_uid
    )
    if ptvs:
        condition += " and roi_name in ('%s')" % "','".join(ptvs)
    ptv_coordinates_strings = query("dvhs", "roi_coord_string", condition)

    ptvs = [
        roi_form.get_planes_from_string(ptv[0])
        for ptv in ptv_coordinates_strings
    ]

    return roi_geom.union(ptvs)


def get_treatment_volume_centroid(tv):
    """Get the centroid of a treatment volume

    Parameters
    ----------
    tv : dict
        treatment volume formatted as a "sets of points" object specified in
        tools.roi_geometry

    Returns
    -------
    np.ndarray
        DICOM coordinates of treatment volume centroid

    """
    return np.array(roi_geom.centroid(tv))


def get_treatment_volume_coord(tv):
    """Get the volume

    Parameters
    ----------
    tv : dict
        treatment volume formatted as a "sets of points" object specified in
        tools.roi_geometry

    Returns
    -------
    list
        The coordinates of the treatment volume, using
        tools.roi_formatter.get_roi_coordinates_from_planes

    """
    return roi_form.get_roi_coordinates_from_planes(tv)


def query(table, column, condition, unique=False):
    """Helper function, automatically creates connection for query

    Parameters
    ----------
    table : str
        DVHs', 'Plans', 'Rxs', 'Beams', or 'DICOM_Files'
    column : str: str
        a csv of SQL columns to be returned
    condition : str: str
        a condition in SQL syntax
    unique : bool, optional
        Call DVH_SQL.get_unique_values if true, DVH_SQL.query if false

    Returns
    -------
    list
        return of query

    """
    with DVH_SQL() as cnx:
        func = cnx.get_unique_values if unique else cnx.query
        ans = func(table, column, condition)
    return ans


def uid_has_ptvs(study_instance_uid):
    """Check if study instance uid has PTVs

    Parameters
    ----------
    study_instance_uid : str
        study instance uid

    Returns
    -------
    bool
        True if uid has any roi_name with an roi_type LIKE PTV%

    """
    with DVH_SQL() as cnx:
        condition = (
            "study_instance_uid = '%s' and roi_type LIKE 'PTV%%'"
            % study_instance_uid
        )
        ans = cnx.query("DVHs", "roi_type", condition)
    return bool(ans)


def update_roi_metric(
    roi_metric_calc, uid, callback=None, centroid_calc=False, ptv_calc=True
):
    """Call roi_metric_calc for ``uid``

    Parameters
    ----------
    roi_metric_calc : callable
        Function with parameters: study_insstance_uid, roi_name, precalc
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'
    centroid_calc : bool, optional
        Update pre_calc with get_treatment_volume_centroid
    ptv_calc : bool, optional
        roi_metric_calc involves getting a total treatment volume

    """
    pre_calc = None
    if ptv_calc:
        condition = "study_instance_uid = '%s' and roi_type like 'PTV%%'" % uid
        ptvs = [row[0] for row in query("DVHs", "roi_name", condition)]
        if len(ptvs):
            pre_calc = get_total_treatment_volume_of_study(uid, ptvs=ptvs)
            if centroid_calc:
                pre_calc = get_treatment_volume_centroid(pre_calc)

    if pre_calc is not None or not ptv_calc:
        condition = (
            "study_instance_uid = '%s' and roi_type not like 'PTV%%'" % uid
        )
        rois = [row[0] for row in query("DVHs", "roi_name", condition)]

        for i, roi in enumerate(rois):
            if pre_calc is None:
                roi_metric_calc(uid, roi)
            else:
                roi_metric_calc(uid, roi, pre_calc=pre_calc)
            if callback is not None:
                msg = {
                    "label": "Processing (%s of %s): %s"
                    % (i + 1, len(rois), roi),
                    "gauge": float(i / len(rois)),
                }
                callback(msg)


def update_ptv_dist_data(uid, callback=None):
    """Update PTV distance data with min_distances function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(min_distances, uid, callback)


def update_ptv_overlap(uid, callback=None):
    """Update PTV overlap data with treatment_volume_overlap function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(treatment_volume_overlap, uid, callback)


def update_ptv_centroid_distances(uid, callback=None):
    """Update PTV centroid distances with dist_to_ptv_centroids function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(dist_to_ptv_centroids, uid, callback, centroid_calc=True)


def update_roi_centroid(uid, callback=None):
    """Update ROI centroids with centroid function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(centroid, uid, callback, ptv_calc=False)


def update_roi_spread(uid, callback=None):
    """Update ROI spread values with spread function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(spread, uid, callback, ptv_calc=False)


def update_roi_cross_section(uid, callback=None):
    """Update ROI cross section data with cross_section function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(cross_section, uid, callback, ptv_calc=False)


def update_roi_surface_area(uid, callback=None):
    """Update ROI surface areas with surface_area function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(surface_area, uid, callback, ptv_calc=False)


def update_roi_volume(uid, callback=None):
    """Update ROI volume with volumes function

    Parameters
    ----------
    uid : str
        study instance uid
    callback : callable, optional
        Accepts a dict with keys of 'label' and 'gauge'

    """
    update_roi_metric(volumes, uid, callback, ptv_calc=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.sql_to_python.py
"""Query a DVHA SQL table and parse the return into a python object"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from dvha.db.sql_connector import DVH_SQL
from dateutil.parser import parse as date_parser
from dvha.tools.errors import push_to_log


class QuerySQL:
    """Object to generically query a specified table. Each column is stored as
    a property of the object

    For example, if you query 'dvhs' with condition string of
    "mrn = 'some_mrn'" you can access any column name 'some_column' with
    QuerySQL.some_column which will return a list of values for 'some_column'.
    All properties contain lists with the order of their values synced,
    unless unique=True

    Parameters
    ----------
    table_name : str
        Beams', 'DVHs', 'Plans', or 'Rxs'
    condition_str : str
        condition in SQL syntax
    unique : bool, optional
        If True, only unique values stored
    columns : list, optional
        A list of SQL column names to be included in the return. If left as
        ``None``, all columns will be returned
    group : int, optional
        either 1 or 2

    """

    def __init__(
        self, table_name, condition_str, unique=False, columns=None, group=1
    ):

        table_name = table_name.lower()

        if table_name in {"beams", "dvhs", "plans", "rxs"}:
            self.table_name = table_name
            self.condition_str = condition_str
            with DVH_SQL(group=group) as cnx:

                all_columns = cnx.get_column_names(table_name)
                if columns is not None:
                    columns = set(all_columns).intersection(
                        columns
                    )  # ensure provided columns exist in SQL table
                else:
                    columns = all_columns

                for column in columns:
                    if column not in {
                        "roi_coord_string",
                        "distances_to_ptv",
                    }:  # ignored for memory since not used here
                        self.cursor = cnx.query(
                            self.table_name, column, self.condition_str
                        )
                        force_date = cnx.is_sqlite_column_datetime(
                            self.table_name, column
                        )  # returns False for pgsql
                        rtn_list = self.cursor_to_list(force_date=force_date)
                        if unique:
                            rtn_list = get_unique_list(rtn_list)
                        # create property of QuerySQL based on SQL column name
                        setattr(self, column, rtn_list)
        else:
            push_to_log(
                msg="QuerySQL: Table name in valid. Please select from Beams, "
                "DVHs, Plans, or Rxs."
            )

    def cursor_to_list(self, force_date=False):
        """Convert a cursor return into a list of values

        Parameters
        ----------
        force_date : bool, optional
             Apply dateutil.parser to values

        Returns
        -------
        list
            queried data

        """
        rtn_list = []
        for row in self.cursor:
            if force_date:
                try:
                    if type(row[0]) is int:
                        rtn_list.append(str(date_parser(str(row[0]))))
                    else:
                        rtn_list.append(str(date_parser(row[0])))
                except Exception:
                    rtn_list.append("None")

            elif isinstance(row[0], (int, float)):
                rtn_list.append(row[0])
            else:
                rtn_list.append(str(row[0]))
        return rtn_list


def get_unique_list(input_list):
    """Remove duplicates in list and retain order

    Parameters
    ----------
    input_list : list
        any list of objects

    Returns
    -------
    list
        input_list without duplicates

    """
    rtn_list_unique = []
    for value in input_list:
        if value not in rtn_list_unique:
            rtn_list_unique.append(value)

    return rtn_list_unique


def get_database_tree():
    """Query SQL to get all columns of each table

    Parameters
    ----------

    Returns
    -------
    dict
        column data sorted by table

    """
    with DVH_SQL() as cnx:
        tree = {table: cnx.get_column_names(table) for table in cnx.tables}
    return tree



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.dicom_parser.py
"""Parse DICOM RT Dose, Structure, and Plan files for the DVHA SQL Database"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from dicompylercore import dvhcalc, dvh as dicompyler_dvh
from dicompylercore.dicomparser import DicomParser as dicompylerParser
from datetime import datetime
from dateutil.relativedelta import relativedelta  # python-dateutil
from dateutil.parser import parse as date_parser
import numpy as np
from os.path import basename, join
from pubsub import pub
import pydicom
from dvha.options import Options
from dvha.tools.errors import push_to_log
from dvha.tools.roi_name_manager import clean_name, DatabaseROIs
from dvha.tools.utilities import (
    change_angle_origin,
    calc_stats,
    is_date,
    validate_transfer_syntax_uid,
)
from dvha.tools.roi_formatter import (
    dicompyler_roi_coord_to_db_string,
    get_planes_from_string,
)
from dvha.tools import roi_geometry as roi_calc
from mlca.mlc_analyzer import Beam as mlca
from dvha.db.sql_connector import DVH_SQL


class DICOM_Parser:
    """Parse a set of DICOM files for database import

    Parameters
    ----------
    plan_file : str
        absolute path of DICOM RT Plan file
    structure_file : structure_file: str
        absolute path of DICOM RT Struct file
    dose_file : str
        absolute path of DICOM RT Dose file
    dose_sum_file : str
        DICOM RT Dose file with summed dose
    plan_over_rides : dict
        Values from import GUI to override DICOM file data
    global_plan_over_rides : dict
        Values from import GUI to override DICOM file data
    roi_over_ride : dict
        lookup up for roi name and type overrides
    roi_map : DatabaseROIs
        roi name map
    use_dicom_dvh : bool
        use the DVH stored in DICOM RT-Dose if it exists
    plan_ptvs : list, iterable
        assign specific PTVs for distance calculations

    """

    def __init__(
        self,
        plan_file=None,
        structure_file=None,
        dose_file=None,
        dose_sum_file=None,
        plan_over_rides=None,
        global_plan_over_rides=None,
        roi_over_ride=None,
        roi_map=None,
        use_dicom_dvh=False,
        plan_ptvs=None,
        other_dicom_files=None,
    ):

        self.database_rois = DatabaseROIs() if roi_map is None else roi_map

        options = Options()
        self.import_path = options.IMPORTED_DIR
        self.dvh_bin_max_dose = options.dvh_bin_max_dose
        self.dvh_bin_max_dose_units = options.dvh_bin_max_dose_units
        self.mlca_options = options.MLC_ANALYZER_OPTIONS
        self.get_dvh_kwargs = options.GET_DVH_KWARGS
        self.dvh_small_volume_threshold = options.DVH_SMALL_VOLUME_THRESHOLD
        self.dvh_high_resolution_factor = options.DVH_HIGH_RESOLUTION_FACTOR
        self.dvh_high_resolution_segments_between = (
            options.DVH_HIGH_RESOLUTION_SEGMENTS_BETWEEN
        )

        self.plan_file = plan_file
        self.structure_file = structure_file
        self.dose_file = dose_file
        self.dose_sum_file = dose_sum_file

        self.use_dicom_dvh = use_dicom_dvh

        # store these values when clearing file loaded data
        self.stored_values = {}

        self.rt_data = {key: None for key in ["plan", "structure", "dose"]}
        self.dicompyler_data = {
            key: None for key in ["plan", "structure", "dose"]
        }
        self.dicompyler_rt_plan = None
        self.structure_name_and_type = None
        if self.plan_file:
            self.dicompyler_data["plan"] = dicompylerParser(self.plan_file)
            self.rt_data["plan"] = self.dicompyler_data["plan"].ds
            self.dicompyler_rt_plan = self.dicompyler_data["plan"].GetPlan()
        if self.structure_file:
            self.dicompyler_data["structure"] = dicompylerParser(
                self.structure_file
            )
            self.rt_data["structure"] = self.dicompyler_data["structure"].ds
            self.structure_name_and_type = self.get_structure_name_and_type(
                self.rt_data["structure"]
            )
        if self.dose_file:
            if self.dose_sum_file is None:
                # self.rt_data['dose'] = dicompylerParser(self.dose_file).ds
                # The above line may lead to OSError: Deferred read --
                # original filename not stored. Cannot re-open
                # switching back to pydicom.read_file()
                self.rt_data["dose"] = pydicom.read_file(
                    self.dose_file, force=True
                )
            else:
                self.rt_data["dose"] = pydicom.read_file(
                    self.dose_sum_file, force=True
                )

        # These properties are not inherently stored in Pinnacle DICOM files,
        # but can be extracted from dummy ROI
        # names automatically generated by the Pinnacle Script provided by
        # DVH Analytics
        self.poi_rx_data = self.get_rx_data_from_dummy_rois()
        self.poi_tx_site = self.get_tx_site_from_dummy_rois()

        self.__initialize_rx_beam_and_ref_beam_data()

        # over ride objects
        keys = [
            "mrn",
            "study_instance_uid",
            "birth_date",
            "sim_study_date",
            "physician",
            "tx_site",
            "rx_dose",
        ]
        self.plan_over_rides = [plan_over_rides, {key: None for key in keys}][
            plan_over_rides is None
        ]
        self.global_plan_over_rides = global_plan_over_rides

        self.roi_over_ride = (
            roi_over_ride
            if roi_over_ride is not None
            else {"name": {}, "type": {}}
        )
        for over_ride_type, over_ride in self.roi_over_ride.items():
            for key, value in over_ride.items():
                self.structure_name_and_type[key][over_ride_type] = value

        self.plan_ptvs = plan_ptvs
        self.other_dicom_files = other_dicom_files

    def update_stored_values(self):
        """Update stored_values, meant for PreImportData"""
        keys = [
            "study_instance_uid_to_be_imported",
            "patient_name",
            "mrn",
            "sim_study_date",
            "birth_date",
            "rx_dose",
            "ptv_names",
            "physician",
            "ptv_exists",
            "tx_site",
            "patient_orientation",
        ]
        self.stored_values = {key: getattr(self, key) for key in keys}

    def __initialize_rx_beam_and_ref_beam_data(self):
        beam_num = 0
        self.rx_data = []
        self.beam_data = {}
        self.ref_beam_data = []

        if hasattr(self.rt_data["plan"], "FractionGroupSequence"):
            for fx_grp_index, fx_grp_seq in enumerate(
                self.rt_data["plan"].FractionGroupSequence
            ):
                self.rx_data.append(
                    RxParser(
                        self.rt_data["plan"],
                        self.dicompyler_rt_plan,
                        self.rt_data["structure"],
                        fx_grp_index,
                        self.poi_rx_data,
                        self.study_instance_uid_to_be_imported,
                    )
                )
                self.beam_data[fx_grp_index] = []
                for fx_grp_beam in range(int(fx_grp_seq.NumberOfBeams)):
                    beam_number = self.beam_sequence[beam_num].BeamNumber
                    beam_seq = self.beam_sequence[beam_num]
                    cp_seq = self.get_cp_sequence(self.beam_sequence[beam_num])
                    ref_beam_seq_index = (
                        self.get_referenced_beam_sequence_index(
                            fx_grp_seq, beam_number
                        )
                    )
                    ref_beam_seq = fx_grp_seq.ReferencedBeamSequence[
                        ref_beam_seq_index
                    ]
                    self.beam_data[fx_grp_index].append(
                        BeamParser(
                            beam_seq, ref_beam_seq, cp_seq, self.mlca_options
                        )
                    )

                    beam_num += 1
        else:  # https://github.com/cutright/DVH-Analytics/issues/127
            self.rx_data.append(
                RxParser(
                    self.rt_data["plan"],
                    self.dicompyler_rt_plan,
                    self.rt_data["structure"],
                    0,
                    self.poi_rx_data,
                    self.study_instance_uid_to_be_imported,
                )
            )
            self.beam_data[0] = []
            for beam_num in range(len(self.beam_sequence)):
                beam_seq = self.beam_sequence[beam_num]
                cp_seq = self.get_cp_sequence(self.beam_sequence[beam_num])
                self.beam_data[0].append(
                    BeamParser(beam_seq, None, cp_seq, self.mlca_options)
                )

    @staticmethod
    def get_referenced_beam_sequence_index(fx_grp_seq, beam_number):
        """Determine the index of the referenced beam sequence for a given
        beam_number. This properly BeamSequence beams and
        ReferencedBeamSequence beams.

        Parameters
        ----------
        fx_grp_seq :
            Index of the fraction group (likely 0 unless TPS allows multiple
            prescriptions)
        beam_number :
            The beam number of the beam of interest within BeamSequence

        Returns
        -------
        int
            the ReferencedBeamSequence index that matches beam_number

        """
        for i, ref_beam_seq in enumerate(fx_grp_seq.ReferencedBeamSequence):
            if ref_beam_seq.ReferencedBeamNumber == beam_number:
                return i
        push_to_log(
            msg="DICOM_Parser: Failed to find a matching reference beam in "
            "ReferencedBeamSequence for beam number %s" % beam_number
        )

    def get_rx_data_from_dummy_rois(self):
        """DVH Analytics allows for the user to indicate prescription
        information embedded in POI names. Some TPSs do not include
        prescription information in the exported DICOM Plan file. For example,
        Pinnacle scripts are provided by DVHA which will create these POIs.
        The scripts are found in resources/Pinnacle Scripts

        A POI in the format:
        "rx#: <rx name>: <rx dose in cGy> cGy x <fxs> to <norm. %>%: <norm. method>: <norm. object>"
        where # is rx number starting from 1, will allow DVH-Analytics to
        retrieve rx information.

        Returns
        -------
        dict
            data for the Rxs table that can be extracted from POIs

        """

        if self.structure_file:
            struct_seq = self.rt_data["structure"].StructureSetROISequence
            rx_indices = [
                i
                for i, roi in enumerate(struct_seq)
                if roi.ROIName.lower().startswith("rx ")
            ]

            rx_data = {}
            for i in rx_indices:
                roi_name = struct_seq[i].ROIName.lower()
                name_split = roi_name.split(":")
                if len(name_split) > 3:
                    fx_grp_number = int(name_split[0].strip("rx "))
                    fx_grp_name = name_split[1].strip()
                    fx_dose = float(name_split[2].split("cgy")[0]) / 100.0
                    fxs = int(name_split[2].split("x ")[1].split(" to")[0])
                    rx_dose = fx_dose * float(fxs)
                    rx_percent = float(
                        name_split[2].strip().split(" ")[5].strip("%")
                    )
                    normalization_method = name_split[3].strip()

                    if len(name_split) > 4:
                        normalization_object = [
                            "plan_max",
                            name_split[4].strip(),
                        ][normalization_method != "plan_max"]
                    else:
                        normalization_object = None

                    rx_data[fx_grp_number] = {
                        "fx_grp_number": fx_grp_number,
                        "fx_grp_name": fx_grp_name,
                        "fx_dose": fx_dose,
                        "fxs": fxs,
                        "rx_dose": rx_dose,
                        "rx_percent": rx_percent,
                        "normalization_method": normalization_method,
                        "normalization_object": normalization_object,
                    }

            return rx_data

    def get_tx_site_from_dummy_rois(self):
        """DVH Analytics allows for the user to indicate treatment information
        embedded in POI names. By default, DVHA assumes the plan name is the
        tx site. However, if a POI is provided as follows, the tx site
        provided in the POI will be used instead. Pinnacle scripts are
        provided by DVHA which will create these POIs. The scripts are found
        in resources/Pinnacle Scripts

        A POI in the format:
        "tx: <site>"
        will allow DVH-Analytics to add this site name to the database upon
        import.

        Returns
        -------
        str
            treatment site name

        """

        if self.structure_file:
            struct_seq = self.rt_data["structure"].StructureSetROISequence
            tx_indices = [
                i
                for i, roi in enumerate(struct_seq)
                if roi.ROIName.lower().startswith("tx: ")
            ]

            if tx_indices:
                return struct_seq[tx_indices[0]].ROIName.split("tx: ")[1]

    def get_plan_row(self):
        """Get all data needed for a row in the Plans table of the database

        Returns
        -------
        dict
            plan row data with the column name as the key and the values are
            lists in the format [value, column variable type]. This object
            will be passed to DVH_SQL.insert_row

        """

        return {
            "mrn": [self.mrn, "text"],
            "study_instance_uid": [
                self.study_instance_uid_to_be_imported,
                "text",
            ],
            "birth_date": [self.birth_date, "date"],
            "age": [self.age, "smallint"],
            "patient_sex": [self.patient_sex, "char(1)"],
            "sim_study_date": [self.sim_study_date, "date"],
            "physician": [self.physician, "varchar(50)"],
            "tx_site": [self.tx_site, "varchar(50)"],
            "rx_dose": [self.rx_dose, "real"],
            "fxs": [self.fxs_total, "int"],
            "patient_orientation": [self.patient_orientation, "varchar(3)"],
            "plan_time_stamp": [self.plan_time_stamp, "timestamp"],
            "struct_time_stamp": [self.struct_time_stamp, "timestamp"],
            "dose_time_stamp": [self.dose_time_stamp, "timestamp"],
            "tps_manufacturer": [self.tps_manufacturer, "varchar(50)"],
            "tps_software_name": [self.tps_software_name, "varchar(50)"],
            "tps_software_version": [self.tps_software_version, "varchar(30)"],
            "tx_modality": [self.tx_modality, "varchar(30)"],
            "tx_time": [self.tx_time, "time"],
            "total_mu": [self.total_mu, "real"],
            "dose_grid_res": [self.dose_grid_res, "varchar(16)"],
            "heterogeneity_correction": [
                self.heterogeneity_correction,
                " varchar(30)",
            ],
            "baseline": [None, "boolean"],
            "import_time_stamp": [None, "timestamp"],
            "protocol": [None, "text"],
            "toxicity_grades": [None, "text"],
            "complexity": [self.plan_complexity, "real"],
        }

    def get_beam_rows(self):
        """Get all beam rows across all prescriptions

        Returns
        -------
        list
            a complete list of beam rows

        """
        beam_rows = []
        for rx_index, beam_set in self.beam_data.items():
            for beam in beam_set:
                if float(beam.beam_mu) > 0:  # Ignores set-up fields
                    beam_rows.append(self.get_beam_row(rx_index, beam))

        if len(beam_rows) == 0:
            return self.dummy_beam_row

        return beam_rows

    @property
    def dummy_beam_row(self):
        """Empty beam row object"""
        return [
            {
                "mrn": [self.mrn, "text"],
                "study_instance_uid": [
                    self.study_instance_uid_to_be_imported,
                    "text",
                ],
                "import_time_stamp": [None, "timestamp"],
            }
        ]

    def get_beam_row(self, rx_index, beam):
        """Get all data needed for a row in the Beams table of the database

        Parameters
        ----------
        rx_index : int
            the index of self.rx_data associated with beam
        beam : BeamParser
            the beam data object to be parsed

        Returns
        -------
        dict
            beam row data with the column name as the key and the values are
            lists in the format [value, column variable type]. This object
            will be passed to DVH_SQL.insert_row

        """
        rx = self.rx_data[rx_index]

        # store these getters so code is repeated every reference
        gantry_values = beam.gantry_values
        collimator_values = beam.collimator_values
        couch_values = beam.couch_values
        mlc_stat_data = beam.mlc_stat_data

        return {
            "mrn": [self.mrn, "text"],
            "study_instance_uid": [
                self.study_instance_uid_to_be_imported,
                "text",
            ],
            "beam_number": [beam.beam_number, "int"],
            "beam_name": [beam.beam_name, "varchar(30)"],
            "fx_grp_number": [rx.fx_grp_number, "smallint"],
            "fx_count": [rx.fx_count, "int"],
            "fx_grp_beam_count": [rx.beam_count, "smallint"],
            "beam_dose": [beam.beam_dose, "real"],
            "beam_mu": [beam.beam_mu, "real"],
            "radiation_type": [beam.radiation_type, "varchar(30)"],
            "beam_energy_min": [beam.energy_min, "real"],
            "beam_energy_max": [beam.energy_max, "real"],
            "beam_type": [beam.beam_type, "varchar(30)"],
            "control_point_count": [beam.control_point_count, "int"],
            "gantry_start": [gantry_values["start"], "real"],
            "gantry_end": [gantry_values["end"], "real"],
            "gantry_rot_dir": [gantry_values["rot_dir"], "varchar(5)"],
            "gantry_range": [gantry_values["range"], "real"],
            "gantry_min": [gantry_values["min"], "real"],
            "gantry_max": [gantry_values["max"], "real"],
            "collimator_start": [collimator_values["start"], "real"],
            "collimator_end": [collimator_values["end"], "real"],
            "collimator_rot_dir": [collimator_values["rot_dir"], "varchar(5)"],
            "collimator_range": [collimator_values["range"], "real"],
            "collimator_min": [collimator_values["min"], "real"],
            "collimator_max": [collimator_values["max"], "real"],
            "couch_start": [couch_values["start"], "real"],
            "couch_end": [couch_values["end"], "real"],
            "couch_rot_dir": [couch_values["rot_dir"], "varchar(5)"],
            "couch_range": [couch_values["range"], "real"],
            "couch_min": [couch_values["min"], "real"],
            "couch_max": [couch_values["max"], "real"],
            "beam_dose_pt": [beam.beam_dose_pt, "varchar(35)"],
            "isocenter": [beam.isocenter, "varchar(35)"],
            "ssd": [beam.ssd, "real"],
            "treatment_machine": [beam.treatment_machine, "varchar(30)"],
            "scan_mode": [beam.scan_mode, "varchar(30)"],
            "scan_spot_count": [beam.scan_spot_count, "real"],
            "beam_mu_per_deg": [beam.beam_mu_per_deg, "real"],
            "beam_mu_per_cp": [beam.beam_mu_per_cp, "real"],
            "import_time_stamp": [None, "timestamp"],
            "area_min": [mlc_stat_data["area"][5] / 100.0, "real"],
            "area_mean": [mlc_stat_data["area"][3] / 100.0, "real"],
            "area_median": [mlc_stat_data["area"][2] / 100.0, "real"],
            "area_max": [mlc_stat_data["area"][0] / 100.0, "real"],
            "perim_min": [mlc_stat_data["area"][5] / 10.0, "real"],
            "perim_mean": [mlc_stat_data["area"][3] / 10.0, "real"],
            "perim_median": [mlc_stat_data["area"][2] / 10.0, "real"],
            "perim_max": [mlc_stat_data["area"][0] / 10.0, "real"],
            "x_perim_min": [mlc_stat_data["x_perim"][5] / 10.0, "real"],
            "x_perim_mean": [mlc_stat_data["x_perim"][3] / 10.0, "real"],
            "x_perim_median": [mlc_stat_data["x_perim"][2] / 10.0, "real"],
            "x_perim_max": [mlc_stat_data["x_perim"][0] / 10.0, "real"],
            "y_perim_min": [mlc_stat_data["y_perim"][5] / 10.0, "real"],
            "y_perim_mean": [mlc_stat_data["y_perim"][3] / 10.0, "real"],
            "y_perim_median": [mlc_stat_data["y_perim"][2] / 10.0, "real"],
            "y_perim_max": [mlc_stat_data["y_perim"][0] / 10.0, "real"],
            "complexity_min": [mlc_stat_data["cmp_score"][5], "real"],
            "complexity_mean": [mlc_stat_data["cmp_score"][3], "real"],
            "complexity_median": [mlc_stat_data["cmp_score"][2], "real"],
            "complexity_max": [mlc_stat_data["cmp_score"][0], "real"],
            "cp_mu_min": [mlc_stat_data["cp_mu"][5], "real"],
            "cp_mu_mean": [mlc_stat_data["cp_mu"][3], "real"],
            "cp_mu_median": [mlc_stat_data["cp_mu"][2], "real"],
            "cp_mu_max": [mlc_stat_data["cp_mu"][0], "real"],
            "complexity": [mlc_stat_data["complexity"], "real"],
            "tx_modality": [beam.tx_modality, "varchar(35)"],
        }

    def get_rx_rows(self):
        """Get all rx rows

        Returns
        -------
        list
            a complete list of rx data objects

        """
        return [self.get_rx_row(rx) for rx in self.rx_data]

    def get_rx_row(self, rx):
        """Get all data needed for a row in the Rxs table of the database

        Parameters
        ----------
        rx : RxParser
            a rx data object

        Returns
        -------
        dict
            rx row data with the column name as the key and the values are
            lists in the format [value, column variable type]. This object
            will be passed to DVH_SQL.insert_row

        """

        data = {
            "mrn": [self.mrn, "text"],
            "study_instance_uid": [
                self.study_instance_uid_to_be_imported,
                "text",
            ],
            "plan_name": [rx.plan_name, "varchar(50)"],
            "fx_grp_name": [rx.fx_grp_name, "varchar(30)"],
            "fx_grp_number": [rx.fx_grp_number, "smallint"],
            "fx_grp_count": [self.fx_grp_count, "smallint"],
            "fx_dose": [rx.fx_dose, "real"],
            "fxs": [rx.fx_count, "smallint"],
            "rx_dose": [rx.rx_dose, "real"],
            "rx_percent": [None, "real"],
            "normalization_method": [rx.normalization_method, "varchar(30)"],
            "normalization_object": [rx.normalization_object, "varchar(30)"],
            "import_time_stamp": [None, "timestamp"],
        }

        # over-ride values if dummy roi's are used to store rx data in
        # rt_structure
        if self.poi_rx_data and rx.fx_grp_number in list(self.poi_rx_data):
            for key, value in self.poi_rx_data[rx.fx_grp_number].items():
                data[key][0] = value

        for key, value in self.plan_over_rides.items():
            if key in data and value is not None:
                data[key][0] = value

        return data

    def get_dvh_row(self, dvh_index):
        """Get all data needed for a row in the DVHs table of the database

        Parameters
        ----------
        dvh_index : int
            the index of the ROI to be imported

        Returns
        -------
        dict
            dvh row data with the column name as the key and the values are
            lists in the format [value, column variable type]. This object
            will be passed to DVH_SQL.insert_row

        """

        # dicompyler-core expects integer limit in cGy
        # self.rx_dose in Gy, bin max is either in Gy or % Rx dose
        # This is needed to prevent np.histogram from blowing up memory usage
        # if no rx_dose is found, default to the absolute dose
        if self.rx_dose:
            limit = int(
                self.rx_dose
                * self.dvh_bin_max_dose[self.dvh_bin_max_dose_units]
            )
        else:
            limit = int(self.dvh_bin_max_dose["Gy"] * 100.0)

        dvh = None
        if self.use_dicom_dvh and self.dose_sum_file is None:
            try:
                dvh = dicompyler_dvh.DVH.from_dicom_dvh(
                    self.rt_data["dose"], dvh_index
                )
            # raised if structure is not found in DICOM DVH
            except AttributeError:
                pass

        if dvh is None:
            kwargs = {
                key: value for key, value in self.get_dvh_kwargs.items()
            }  # make copy
            kwargs["structure"] = self.rt_data["structure"]
            kwargs["dose"] = self.rt_data["dose"]
            kwargs["roi"] = dvh_index
            kwargs["limit"] = limit
            kwargs["callback"] = self.send_dvh_progress

            # Remove kwargs invalid before 0.5.6
            from dicompylercore import __version__ as dicompylercore_version

            if dicompylercore_version == "0.5.5":
                if "memmap_rtdose" in kwargs.keys():
                    kwargs.pop("memmap_rtdose")

            try:
                try:
                    dvh = dvhcalc.get_dvh(**kwargs)
                except AttributeError:
                    kwargs["dose"] = validate_transfer_syntax_uid(
                        self.rt_data["dose"]
                    )
                    kwargs["structure"] = validate_transfer_syntax_uid(
                        self.rt_data["structure"]
                    )
                    dvh = dvhcalc.get_dvh(**kwargs)
            except MemoryError:
                kwargs["memmap_rtdose"] = True
                dvh = dvhcalc.get_dvh(**kwargs)

            # If small volume, increase resolution
            if dvh.volume < self.dvh_small_volume_threshold:
                try:
                    kwargs["interpolation_resolution"] = (
                        self.rt_data["dose"].PixelSpacing[0]
                        / self.dvh_high_resolution_factor,
                        self.rt_data["dose"].PixelSpacing[1]
                        / self.dvh_high_resolution_factor,
                    )
                    if dicompylercore_version == "0.5.5":
                        kwargs["interpolation_resolution"] = kwargs[
                            "interpolation_resolution"
                        ][0]

                    kwargs[
                        "interpolation_segments_between_planes"
                    ] = self.dvh_high_resolution_segments_between

                    try:
                        dvh_new = dvhcalc.get_dvh(**kwargs)
                    except MemoryError:
                        kwargs["memmap_rtdose"] = True
                        # dicompyler-core needs to re-parse the dose file
                        dvh_new = dvhcalc.get_dvh(**kwargs)
                    dvh = dvh_new
                except Exception as e:
                    msg = (
                        "Small volume calculation failed, "
                        "using default calculation"
                    )
                    push_to_log(e, msg=msg)

        if dvh and dvh.volume > 0:  # ignore points and empty ROIs
            geometries = self.get_dvh_geometries(dvh_index)
            iso_dist = self.get_roi_centroid_to_isocenter_dist(
                geometries["centroid"]
            )

            return {
                "mrn": [self.mrn, "text"],
                "study_instance_uid": [
                    self.study_instance_uid_to_be_imported,
                    "text",
                ],
                "institutional_roi": [
                    self.get_institutional_roi(dvh_index),
                    "varchar(50)",
                ],
                "physician_roi": [
                    self.get_physician_roi(dvh_index),
                    "varchar(50)",
                ],
                "roi_name": [self.get_roi_name(dvh_index), "varchar(50)"],
                "roi_type": [self.get_roi_type(dvh_index), "varchar(20)"],
                "volume": [dvh.volume, "real"],
                "min_dose": [dvh.min, "real"],
                "mean_dose": [dvh.mean, "real"],
                "max_dose": [dvh.max, "real"],
                "dvh_string": [
                    ",".join(["%.5f" % num for num in dvh.counts]),
                    "text",
                ],
                "roi_coord_string": [geometries["roi_coord_str"], "text"],
                "dist_to_ptv_min": [None, "real"],
                "dist_to_ptv_mean": [None, "real"],
                "dist_to_ptv_median": [None, "real"],
                "dist_to_ptv_max": [None, "real"],
                "surface_area": [geometries["surface_area"], "real"],
                "ptv_overlap": [None, "real"],
                "import_time_stamp": [None, "timestamp"],
                "centroid": [
                    ",".join(
                        [str(round(x, 3)) for x in geometries["centroid"]]
                    ),
                    "varchar(35)",
                ],
                "dist_to_ptv_centroids": [None, "real"],
                "dth_string": [None, "text"],
                "spread_x": [geometries["spread"][0], "real"],
                "spread_y": [geometries["spread"][1], "real"],
                "spread_z": [geometries["spread"][2], "real"],
                "cross_section_max": [
                    geometries["cross_sections"]["max"],
                    "real",
                ],
                "cross_section_median": [
                    geometries["cross_sections"]["median"],
                    "real",
                ],
                "centroid_dist_to_iso_min": [iso_dist["min"], "real"],
                "centroid_dist_to_iso_max": [iso_dist["max"], "real"],
                "toxicity_grade": [None, "smallint"],
            }

    def get_roi_centroid_to_isocenter_dist(self, roi_centroid):
        """Get the distance from ROI centroid to beam isocenter

        Parameters
        ----------
        roi_centroid : list
            centroid of ROI in DICOM coordinates

        Returns
        -------
        dict
            min and max iso-to-centroid distances (checked over every beam)

        """
        centroid = np.array(roi_centroid)
        distances = []
        for beam_set in self.beam_data.values():
            for beam in beam_set:
                try:
                    distances.append(
                        np.linalg.norm(beam.isocenter_np - centroid)
                    )
                except Exception as e:
                    msg = (
                        "ROI Centroid calculation failed for mrn: %s, "
                        "could not parse beam isocenter" % self.mrn
                    )
                    push_to_log(e, msg=msg)
        if distances:
            return {
                "min": float(np.min(distances) / 10),
                "max": float(np.max(distances)) / 10,
            }
        return {"min": None, "max": None}

    def get_dicom_file_row(self):
        """Get all data needed for a row in the DICOM_Files table of the
        database

        Returns
        -------
        dict
            dicom file data with the column name as the key and the values are
            lists in the format [value, column variable type]. This object
            will be passed to DVH_SQL.insert_row

        """
        return {
            "mrn": [self.mrn, "text"],
            "study_instance_uid": [
                self.study_instance_uid_to_be_imported,
                "text",
            ],
            "folder_path": [str(join(self.import_path, self.mrn)), "text"],
            "plan_file": [basename(self.plan_file), "text"],
            "structure_file": [basename(self.structure_file), "text"],
            "dose_file": [basename(self.dose_file), "text"],
            "import_time_stamp": [None, "timestamp"],
        }

    @property
    def mrn(self):
        """medical record number

        Returns
        -------
        str
            mrn, with over rides applied
        """
        if self.plan_over_rides["mrn"] is not None:
            return self.plan_over_rides["mrn"]
        return self.rt_data["plan"].PatientID

    @property
    def study_instance_uid(self):
        """study instance uid from DICOM

        Returns
        -------
        str
            study instance uid from DICOM-RT Plan
        """
        return self.rt_data["plan"].StudyInstanceUID

    @property
    def study_instance_uid_to_be_imported(self):
        """study instance uid to be written to database

        Returns
        -------
        str
            study instance uid
        """
        if (
            hasattr(self, "plan_over_rides")
            and self.plan_over_rides["study_instance_uid"]
        ):
            return self.plan_over_rides["study_instance_uid"]
        return self.rt_data["plan"].StudyInstanceUID

    @property
    def is_study_instance_uid_valid(self):
        """Check if uid is used in database

        Returns
        -------
        bool
            False if study instance uid from DICOM is used in database
        """
        return self.is_uid_valid(self.study_instance_uid)

    @property
    def is_study_instance_uid_to_be_imported_valid(self):
        """Check if uid is used in database

        Returns
        -------
        bool
            False if study instance uid from DICOM is used in database
        """
        return self.is_uid_valid(self.study_instance_uid_to_be_imported)

    @staticmethod
    def is_uid_valid(uid):
        """Create DB connection, check if uid exists in database

        Parameters
        ----------
        uid : str
            study instance uid

        Returns
        -------
        bool
            True if uid is used in the database
        """
        valid = False
        if uid:
            with DVH_SQL() as cnx:
                valid = not cnx.is_uid_imported(uid)
        return valid

    @property
    def patient_name(self):
        """PatientName from DICOM-RT Plan

        Returns
        -------
        str
            PatientName (0010, 0010)
        """
        return str(self.rt_data["plan"].PatientName)

    @property
    def file_set_complete(self):
        """Check if plan, structure, and dose files have been set

        Returns
        -------
        bool
            True if plan, structure, and dose files are set
        """
        return all([self.plan_file, self.structure_file, self.dose_file])

    @property
    def missing_files(self):
        """Get missing DICOM file types

        Returns
        -------
        list
            values of 'PLAN', 'STRUCTURE', 'DOSE' if missing
        """
        return [
            key.capitalize()
            for key, value in self.rt_data.items()
            if value is None
        ]

    # ------------------------------------------------------------------------
    # Plan table data
    # ------------------------------------------------------------------------
    @property
    def tx_modality(self):
        """Treatment modality

        Returns
        -------
        str
            CSV of BeamParser.tx_modality values
        """
        tx_modalities = []
        for fx_grp_index, beam_parser_list in self.beam_data.items():
            for beam_parser in beam_parser_list:
                tx_modalities.append(beam_parser.tx_modality)
        tx_modalities = sorted(list(set(tx_modalities)))
        return ",".join(tx_modalities)

    @property
    def rx_dose(self):
        """Prescription dose

        Returns
        -------
        int, float
            Prescription dose in Gy, summed over all Rx rows
        """
        if self.plan_over_rides["rx_dose"] is not None:
            ans = self.plan_over_rides["rx_dose"]
        elif self.poi_rx_data:
            ans = sum([rx["rx_dose"] for rx in self.poi_rx_data.values()])
        else:
            ans = sum([rx.rx_dose for rx in self.rx_data if rx.rx_dose])
        return self.process_global_over_ride("rx_dose", ans)

    @property
    def total_mu(self):
        """Total MU delivered

        Returns
        -------
        int, float
            The total MUs over all Rxs and fractions
        """
        mus = []
        for fx_grp_index, beam_parser_list in self.beam_data.items():
            for beam_parser in beam_parser_list:
                if self.rx_data[fx_grp_index].fx_count:
                    if beam_parser.beam_mu:
                        mus.append(
                            beam_parser.beam_mu
                            * self.rx_data[fx_grp_index].fx_count
                        )
        return sum(mus)

    @property
    def heterogeneity_correction(self):
        """DICOM-RT Dose TissueHeterogeneityCorrection

        Returns
        -------
        str
            TissueHeterogeneityCorrection (3004,0014)

        """
        heterogeneity_correction = "IMAGE"
        try:
            if hasattr(self.rt_data["dose"], "TissueHeterogeneityCorrection"):
                if isinstance(
                    self.rt_data["dose"].TissueHeterogeneityCorrection, str
                ):
                    heterogeneity_correction = self.rt_data[
                        "dose"
                    ].TissueHeterogeneityCorrection
                else:
                    heterogeneity_correction = ",".join(
                        self.rt_data["dose"].TissueHeterogeneityCorrection
                    )
        except Exception as e:
            msg = "Could not extract heterogeneity correction."
            push_to_log(e, msg=msg)
            heterogeneity_correction = ""

        return heterogeneity_correction

    @property
    def patient_sex(self):
        """DICOM-RT Plan PatientSex

        Returns
        -------
        str
            PatientSex (0010,0040)
        """
        return self.get_attribute("plan", "PatientSex")

    @property
    def sim_study_date(self):
        """Simulation study date

        Returns
        -------
        str
            StudyDate (0008,0020)
        """
        if self.plan_over_rides["sim_study_date"] is not None:
            ans = self.plan_over_rides["sim_study_date"]
        else:
            ans = self.get_attribute("plan", "StudyDate")
            if not ans:
                ans = self.find_first_value_from_other_dicom_files("StudyDate")
                ans = ans if ans else ""
        return self.process_global_over_ride("sim_study_date", ans)

    @property
    def birth_date(self):
        """Birth date from DICOM-RT Plan

        Returns
        -------
        str
            PatientBirthDate (0010,0030)
        """
        if self.plan_over_rides["birth_date"] is not None:
            ans = self.plan_over_rides["birth_date"]
        else:
            ans = self.get_attribute("plan", "PatientBirthDate")
        return self.process_global_over_ride("birth_date", ans)

    @property
    def age(self):
        """Patient age at time of study date

        Returns
        -------
        int
            Patient age (years)
        """
        if self.sim_study_date and self.birth_date:
            try:
                dates = {"sim_study_date": None, "birth_date": None}
                for date_type in list(dates):
                    dates[date_type] = getattr(self, date_type)
                    if type(dates[date_type]) is int or float:
                        dates[date_type] = str(dates[date_type])
                    if type(dates[date_type]) is str:
                        dates[date_type] = date_parser(dates[date_type]).date()
                age = relativedelta(
                    dates["sim_study_date"], dates["birth_date"]
                ).years
                if age >= 0:
                    return age
            except Exception:
                pass

    @property
    def physician(self):
        """Physician from DICOM-RT Plan

        Returns
        -------
        str
            PhysiciansOfRecord (0008,1048), ReferringPhysicianName
            (0008,0090), or 'DEFAULT'
        """
        if self.plan_over_rides["physician"] is not None:
            ans = self.plan_over_rides["physician"]
        else:
            ans = str(
                self.get_attribute(
                    "plan", ["PhysiciansOfRecord", "ReferringPhysicianName"]
                )
            ).upper()
            if not ans:
                ans = "DEFAULT"
        return self.process_global_over_ride("physician", ans)

    @property
    def fxs(self):
        """List of NumberOfFractionsPlanned

        Returns
        -------
        list
            NumberOfFractionsPlanned (300A,0078) for each
            FractionGroupSequence (300A,0070)
        """
        if hasattr(self.rt_data["plan"], "FractionGroupSequence"):
            try:
                fx_grp_seq = self.rt_data["plan"].FractionGroupSequence
                return [
                    int(float(fx_grp.NumberOfFractionsPlanned))
                    for fx_grp in fx_grp_seq
                ]
            except ValueError:
                pass

    @property
    def fxs_total(self):
        """Total number of fractions across fraction groups

        Returns
        -------
        int
            Sum over fxs
        """
        fxs = self.fxs
        if fxs:
            return sum(fxs)

    @property
    def fx_grp_count(self):
        """Number of fraction groups

        Returns
        -------
        int
            Length of FractionGroupSequence (300A,0070)
        """
        if hasattr(self.rt_data["plan"], "FractionGroupSequence"):
            return len(self.rt_data["plan"].FractionGroupSequence)
        return 1

    @property
    def patient_orientation(self):
        """Patient Orientation as stored in DICOM-RT Plan

        Returns
        -------
        str
            PatientSetupSequence (300A,0180)
        """
        # TODO: database assumes only one orientation (i.e., three characters)
        seq = self.get_attribute("plan", "PatientSetupSequence")
        if seq is not None:
            # return ','.join([setup.PatientPosition for setup in seq])
            ans = str(seq[0].PatientPosition)
            return ans if len(ans) < 4 else ans[:3]

    @property
    def plan_time_stamp(self):
        """Datetime object of DICOM-RT Plan

        Returns
        -------
        datetime
            RTPlanDate (300A,0006) and RTPlanTime (300A,0007)
        """
        return self.get_time_stamp("plan")

    @property
    def struct_time_stamp(self):
        """Datetime object of DICOM-RT Structure

        Returns
        -------
        datetime
            StructureSetDate (3006,0008) and StructureSetTime (3006,0009)
        """
        return self.get_time_stamp("structure")

    @property
    def dose_time_stamp(self):
        """Datetime object of DICOM-RT Dose

        Returns
        -------
        datetime
            InstanceCreationDate (0008,0012) and
            InstanceCreationTime (0008,0013)
        """
        return self.get_time_stamp("dose")

    @property
    def tps_manufacturer(self):
        """TPS Manufacturer from DICOM-RT Plan

        Returns
        -------
        str
            Manufacturer (0008,0070)
        """
        return self.get_attribute("plan", "Manufacturer")

    @property
    def tps_software_name(self):
        """TPS name from DICOM-RT Plan

        Returns
        -------
        str
            ManufacturerModelName (0008,1090)
        """
        return self.get_attribute("plan", "ManufacturerModelName")

    @property
    def tps_software_version(self):
        """TPS software version

        Returns
        -------
        str
            SoftwareVersions (0018,1020)
        """
        # Some TPSs may store the version in RT Dose rather than plan
        # SoftwareVersions may also be stored as a string rather than a list
        for rt_type in [
            "plan",
            "dose",
            "structure",
        ]:  # Check each rt_type until SoftwareVersions is found
            version = self.get_attribute(rt_type, "SoftwareVersions")
            if version is not None:
                if isinstance(version, str):
                    version = [version]
                return ",".join(version)

    @property
    def tx_site(self):
        """Treatment site

        Returns
        -------
        str
            RTPlanLabel (300A,0002)
        """
        if self.plan_over_rides["tx_site"] is not None:
            ans = self.plan_over_rides["tx_site"]
        elif self.poi_tx_site is not None:
            ans = self.poi_tx_site
        else:
            ans = self.get_attribute("plan", "RTPlanLabel")
        return self.process_global_over_ride("tx_site", ans)

    @property
    def brachy(self):
        """Check if plan is a brachytherapy plan

        Returns
        -------
        bool
            True if DICOM-RT Plan has BrachyTreatmentType (300A, 0202)
        """
        return hasattr(self.rt_data["plan"], "BrachyTreatmentType")

    @property
    def brachy_type(self):
        """Get the BrachyTreatmentType from DICOM-RT Plan

        Returns
        -------
        str
            BrachyTreatmentType (300A, 0202)
        """
        return self.get_attribute("plan", "BrachyTreatmentType")

    @property
    def proton(self):
        """Check if plan is for an IonBeam

        Returns
        -------
        bool
            True if DICOM-RT Plan has IonBeamSequence (300A, 03A2)
        """
        return hasattr(self.rt_data["plan"], "IonBeamSequence")

    @property
    def beam_sequence(self):
        """Get the beam sequence from DICOM-RT Plan

        Returns
        -------
        list
            Either BeamSequence (300A,0080) or IonBeamSequence (300A, 03A2)
        """
        if hasattr(self.rt_data["plan"], "BeamSequence"):
            return self.rt_data["plan"].BeamSequence
        elif hasattr(self.rt_data["plan"], "IonBeamSequence"):
            return self.rt_data["plan"].IonBeamSequence

    def get_cp_sequence(self, beam_seq):
        """Get the control point sequence for a beam

        Parameters
        ----------
        beam_seq : pydicom beam seq element
            Results from beam_sequence property

        Returns
        -------
        list
            Either ControlPointSequence (300A,0111) or
            IonBeamSequence (300A, 03A8)

        """
        if hasattr(beam_seq, "ControlPointSequence"):
            return beam_seq.ControlPointSequence
        elif hasattr(beam_seq, "IonControlPointSequence"):
            return beam_seq.IonControlPointSequence

    @property
    def photon(self):
        """Check if any beam is a photon beam

        Returns
        -------
        bool
            Check all beams for RadiationType (300A,00C6) of photon
        """
        return self.is_beam_with_rad_type_in_plan("photon")

    @property
    def electron(self):
        """Check if any beam is a photon beam

        Returns
        -------
        bool
            Check all beams for RadiationType (300A,00C6) of photon
        """
        return self.is_beam_with_rad_type_in_plan("electron")

    @property
    def radiation_type(self):
        """Get the radiation types used in this DICOM-RT Plan

        Returns
        -------
        str
            CSV of all radiation types
        """
        rad_types = {
            "PHOTONS": self.photon,
            "ELECTRONS": self.electron,
            "PROTONS": self.proton,
            "BRACHY": self.brachy_type,
        }
        types = sorted(
            [
                rad_type
                for rad_type, rad_value in rad_types.items()
                if rad_value
            ]
        )
        return ",".join(types)

    @property
    def tx_time(self):
        """Get the treatment time (brachy only)

        Returns
        -------
        str
            Summation of ChannelTotalTime (300A,0286), formmated into a
            hr:min:sec str
        """
        if hasattr(self.rt_data["plan"], "BrachyTreatmentType") and hasattr(
            self.rt_data["plan"], "ApplicationSetupSequence"
        ):
            seconds = 0
            for app_seq in self.rt_data["plan"].ApplicationSetupSequence:
                for chan_seq in app_seq.ChannelSequence:
                    seconds += chan_seq.ChannelTotalTime
            m, s = divmod(seconds, 60)
            h, m = divmod(m, 60)
            return "%02d:%02d:%02d" % (h, m, s)

    @property
    def dose_grid_res(self):
        """Dose grid resolution from DICOM-RT Dose

        str
            csv of x,y,z resolution using PixelSpacing (0028,0030) and
            SliceThickness (0018,0050), each rounded to nearest 0.1mm
        """
        try:
            dose_grid_resolution = [
                str(round(float(self.rt_data["dose"].PixelSpacing[0]), 1)),
                str(round(float(self.rt_data["dose"].PixelSpacing[1]), 1)),
            ]
            if (
                hasattr(self.rt_data["dose"], "SliceThickness")
                and self.rt_data["dose"].SliceThickness
            ):
                dose_grid_resolution.append(
                    str(round(float(self.rt_data["dose"].SliceThickness), 1))
                )
            return ", ".join(dose_grid_resolution)
        except Exception:
            pass

    def process_global_over_ride(self, key, pre_over_ride_value):
        """Get the value for ``key``, with over ride applied if applicable

        Parameters
        ----------
        key : str
            Key of DICOM_Parser.global_plan_over_rides
        pre_over_ride_value : any
            Original value of interest

        Returns
        -------
        any
            Apply DICOM_Parser.global_plan_over_rides to the provided key,
            return original value if over ride not applicable

        """
        if self.global_plan_over_rides:
            over_ride = self.global_plan_over_rides[key]
            if over_ride["value"]:
                if not over_ride["only_if_missing"] or (
                    over_ride["only_if_missing"] and not pre_over_ride_value
                ):
                    return over_ride["value"]
        return pre_over_ride_value

    @property
    def plan_complexity(self):
        """Get the sum of beam complexities

        Returns
        -------
        float
            Sum of beam complexities across fx groups (fx weighted) using
            dvha.tools.mlc_analyzer
        """
        plan_complexity = 0.0
        fx_counts = self.fxs

        if fx_counts is None:  # No ReferencedBeamSequence, assume 1 FxGrp
            fx_counts = [1]

        for fx_index, fx in self.beam_data.items():
            fxs = fx_counts[fx_index]
            for beam in fx:
                complexity = beam.mlc_stat_data["complexity"]
                if complexity:
                    plan_complexity += complexity * fxs
        if plan_complexity:
            return plan_complexity / sum(fx_counts)

    # -------------------------------------------------------------------------
    # DVH table data
    # -------------------------------------------------------------------------
    def get_roi_type(self, key):
        """Get the ROI type as defined in DICOM (e.g., PTV, ORGAN,
        EXTERNAL, etc.)

        Parameters
        ----------
        key : int
            index of ROI

        Returns
        -------
        str
            the roi type

        """
        if key in list(self.roi_over_ride["type"]):
            return self.roi_over_ride["type"][key]

        roi_type_from_roi_map = self.database_rois.get_roi_type(
            self.physician, self.get_physician_roi(key)
        )
        if roi_type_from_roi_map != "NONE":
            return roi_type_from_roi_map

        return str(self.structure_name_and_type[key]["type"]).upper()

    def reset_roi_type_over_ride(self, key):
        """Set the roi_type over ride to ``None``

        Parameters
        ----------
        key : int
            index of ROI

        """
        self.roi_over_ride["type"][key] = None

    def get_roi_name(self, key):
        """

        Parameters
        ----------
        key :
            the index of the roi

        Returns
        -------
        str
            roi name to be used in the database

        """
        if key in list(self.roi_over_ride["name"]):
            return clean_name(self.roi_over_ride["name"][key])
        return clean_name(self.structure_name_and_type[key]["name"])

    def get_physician_roi(self, key):
        """Look up the physician roi of the specified roi for this plan

        Parameters
        ----------
        key :
            the index of the roi

        Returns
        -------
        str
            physician roi or 'uncategorized'

        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_physician_roi(
                    self.physician, roi_name
                )
        return "uncategorized"

    def get_institutional_roi(self, key):
        """Look up the institutional roi of the specified roi for this plan

        Parameters
        ----------
        key :
            the index of the roi

        Returns
        -------
        str
            institutional roi or 'uncategorized'

        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_institutional_roi(
                    self.physician, self.get_physician_roi(key)
                )
            if roi_name in self.database_rois.institutional_rois:
                return roi_name
        return "uncategorized"

    @property
    def ptv_exists(self):
        """Check if the plan has at least one PTV assigned

        Returns
        -------
        bool
            True if any roi is of type 'PTV'
        """
        if self.structure_name_and_type:
            for key in list(self.structure_name_and_type):
                if self.get_roi_type(key) == "PTV":
                    return True
        return False

    @property
    def roi_names(self):
        """Get all of the roi names

        Returns
        -------
        list
            roi_names for this plan per the plan (clean_name not applied)

        """
        return [
            self.get_roi_name(key)
            for key in list(self.structure_name_and_type)
        ]

    def get_roi_key(self, roi_name):
        """Look-up the roi index for a given roi_name

        Parameters
        ----------
        roi_name :
            the ROI name to be looked-up

        Returns
        -------
        int
            the roi index or None

        """
        roi_name = clean_name(roi_name)
        for key in list(self.structure_name_and_type):
            if roi_name == clean_name(self.get_roi_name(key)):
                return key

    @property
    def ptv_names(self):
        """Scan all ROIs for PTVs

        Parameters
        ----------

        Returns
        -------
        list
            PTV names

        """
        if self.structure_name_and_type:
            return [
                self.get_roi_name(x)
                for x in list(self.structure_name_and_type)
                if self.get_roi_type(x) == "PTV"
            ]
        return self.stored_values["ptv_names"]

    def get_surface_area(self, key):
        """Get the surface of area of an roi by key

        Parameters
        ----------
        key :
            the index of the roi

        Returns
        -------
        float
            the surface area of the ROI in cm^2

        """
        coord = self.dicompyler_data["structure"].GetStructureCoordinates(key)
        try:
            return roi_calc.surface_area(coord)
        except Exception as e:
            msg = (
                "DICOM_Parser: Surface area calculation failed for key, "
                "name: %s, %s" % (key, self.get_roi_name(key))
            )
            push_to_log(e, msg=msg)

    def get_dvh_geometries(self, key):
        """Collect ROI geometry information to be stored in the database

        Parameters
        ----------
        key :
            the index of the roi

        Returns
        -------
        dict
            geometric information with keys: 'roi_coord_str', 'surface_area',
            'centroid', 'spread', 'cross_sections'

        """

        structure_coord = self.dicompyler_data[
            "structure"
        ].GetStructureCoordinates(key)

        try:
            surface_area = roi_calc.surface_area(structure_coord)
        except Exception as e:
            msg = (
                "DICOM_Parser: Surface area calculation failed for key, "
                "name: %s, %s" % (key, self.get_roi_name(key))
            )
            push_to_log(e, msg=msg)
            surface_area = None

        roi_coord_str = dicompyler_roi_coord_to_db_string(structure_coord)
        planes = get_planes_from_string(roi_coord_str)
        centroid = roi_calc.centroid(planes)
        spread = roi_calc.spread(planes)
        cross_sections = roi_calc.cross_section(planes)

        return {
            "roi_coord_str": roi_coord_str,
            "surface_area": surface_area,
            "centroid": centroid,
            "spread": spread,
            "cross_sections": cross_sections,
        }

    # ------------------------------------------------------------------------
    # Generic tools
    # ------------------------------------------------------------------------
    @property
    def init_param(self):
        """Get initial parameters

        Returns
        -------
        dict
            Extract 'plan_file', 'structure_file', 'dose_file',
            'dose_sum_file', 'plan_over_rides', 'global_plan_over_rides', and
            'roi_over_ride' properties from DICOM_Parser mapped to a dict
        """
        params = [
            "plan_file",
            "structure_file",
            "dose_file",
            "dose_sum_file",
            "plan_over_rides",
            "global_plan_over_rides",
            "roi_over_ride",
        ]
        return {key: getattr(self, key) for key in params}

    def get_attribute(self, rt_type, pydicom_attribute):
        """Short-hand function to get a pydicom attribute

        Parameters
        ----------
        rt_type : str
            plan, dose, or structure
        pydicom_attribute : str or list of str
            attribute as specified in pydicom

        Returns
        -------
        type
            pydicom value or None

        """
        if isinstance(pydicom_attribute, str):
            pydicom_attribute = [pydicom_attribute]

        for attribute in pydicom_attribute:
            if hasattr(self.rt_data[rt_type], attribute):
                return getattr(self.rt_data[rt_type], attribute)

    def get_time_stamp(self, rt_type):
        """Get the time stamp as specified in the DICOM file

        Parameters
        ----------
        rt_type :
            'plan', 'structure' or 'dose'

        Returns
        -------
        type
            datetime

        """

        attribute = {
            "date": {
                "plan": "RTPlanDate",
                "structure": "StructureSetDate",
                "dose": "InstanceCreationDate",
            },
            "time": {
                "plan": "RTPlanTime",
                "structure": "StructureSetTime",
                "dose": "InstanceCreationTime",
            },
        }

        datetime_str = self.get_attribute(rt_type, attribute["date"][rt_type])
        time = self.get_attribute(rt_type, attribute["time"][rt_type])

        try:
            if time:
                time = time.split(".")[0]  # ignore fractional sec
                return datetime.strptime(datetime_str + time, "%Y%m%d%H%M%S")
            else:
                return datetime.strptime(datetime_str, "%Y%m%d")
        except Exception as e:
            push_to_log(e, msg="DICOM_Parser: get_time_stamp failed")

    def is_beam_with_rad_type_in_plan(self, rad_type):
        """If a beam is photon or electron, beam data is stored in
        BeamSequence as opposed to IonBeamSequence

        Parameters
        ----------
        rad_type : str
            the radiation type of the beam (e.g., 'photon' or 'electron')

        Returns
        -------
        bool
            True if any beam in plan is the specified rad_type

        """
        if hasattr(self.rt_data["plan"], "BeamSequence"):
            for beam in self.rt_data["plan"].BeamSequence:
                if rad_type in beam.RadiationType:
                    return True
        return False

    @property
    def pre_import_data(self):
        """Collect data for the Import DICOM window

        Returns
        -------
        dict
            Only the data needed for the pre import screen, pass to
            PreImportData. Designed so actual DICOM files don't need to hang
            out in memory during GUI importer

        """
        self.update_stored_values()
        return {
            "file_set": {
                "plan": self.plan_file,
                "dose": self.dose_file,
                "structure": self.structure_file,
            },
            "stored_values": self.stored_values,
            "dicompyler_rt_structures": self.structure_name_and_type,
            "roi_map": self.database_rois,
        }

    @staticmethod
    def get_structure_name_and_type(rt_struct_ds):
        """Stripped down version of dicompylercore's GetStructures

        Parameters
        ----------
        rt_struct_ds : dicompylercore.dicomparser.DicomParser.ds


        Returns
        -------
        dict
            structure data with keys of ROINumber, containing dicts with keys
            of 'id', 'name', 'type'

        """
        structures = {}

        # Locate the name and number of each ROI
        if "StructureSetROISequence" in rt_struct_ds:
            for item in rt_struct_ds.StructureSetROISequence:
                data = {}
                number = int(item.ROINumber)
                data["id"] = number
                data["name"] = item.ROIName
                structures[number] = data

        # Determine the type of each structure (PTV, organ, external, etc)
        if "RTROIObservationsSequence" in rt_struct_ds:
            for item in rt_struct_ds.RTROIObservationsSequence:
                number = item.ReferencedROINumber
                if number in structures:
                    structures[number]["type"] = item.RTROIInterpretedType

        return structures

    @staticmethod
    def send_dvh_progress(current_plane, plane_count):
        """Callback for dicompyler-core dvh calculation

        Parameters
        ----------
        current_plane : int
            the plane to be calculated
        plane_count : int
            total number of planes to be calculated

        """
        progress = float(current_plane) / float(plane_count)
        pub.sendMessage("update_dvh_progress", msg=progress)

    def find_first_value_from_other_dicom_files(self, attr):
        """Scan ``other_dicom_files`` for CT, MR modality, report back
        value for ``attr``

        Parameters
        ----------
        attr : str
            pydicom attribute

        Returns
        -------
        any
            Lookfs for the first CT or MR DICOM file for the ``attr``. This is
            used, for example, to find a StudyDate when it is missing from
            DICOM-RT Plan

        """
        if (
            self.other_dicom_files
            and self.study_instance_uid in self.other_dicom_files
        ):
            for f in self.other_dicom_files[self.study_instance_uid]:
                ds = pydicom.read_file(f)
                modality = getattr(ds, "Modality")
                if modality.lower() in ["ct", "mr"] and hasattr(ds, attr):
                    return getattr(ds, attr)


class BeamParser:
    """This class is used to parse beam data needed for importing a plan into
    the database

    Parameters
    ----------
    beam_data : BeamSequence element
    ref_beam_data : ReferencedBeamSequence element
    cp_seq : ControlPointSequence
    mlc_analyzer_options : dvha.options.MLC_ANALYZER_OPTIONS

    """

    def __init__(self, beam_data, ref_beam_data, cp_seq, mlc_analyzer_options):

        self.beam_data = beam_data
        self.ref_beam_data = ref_beam_data
        self.cp_seq = cp_seq
        self.options = mlc_analyzer_options

    ##########################################################################
    # General beam parser tools
    ##########################################################################
    @staticmethod
    def get_data_attribute(
        data_obj, pydicom_attr, default=None, data_type=None
    ):
        """Generic method to get the value of a pydicom dataset attribute

        Parameters
        ----------
        data_obj : pydicom dataset
            a pydicom dataset
        pydicom_attr : str
            the attribute of interest
        default : any
            if attribute is not found, this value is returned
        data_type : str
            set to 'float' or 'int' to convert return value as specified

        Returns
        -------
        any
            the value of the pydicom attribute

        """
        if hasattr(data_obj, pydicom_attr):
            value = getattr(data_obj, pydicom_attr)
            if data_type == "float":
                return float(value)
            elif data_type == "int":
                return int(float(value))
            return value
        return default

    def get_point_attribute(self, data_obj, pydicom_attr):
        """Generic method to get the value of a pydicom dataset attribute that
        represents a point

        Parameters
        ----------
        data_obj : pydicom dataset
            a pydicom dataset
        pydicom_attr : str
            the attribute of interest

        Returns
        -------
        str
            a csv of the point values

        """
        point = self.get_data_attribute(data_obj, pydicom_attr)
        if point:
            return ",".join([str(round(dim_value, 3)) for dim_value in point])

    def get_cp_attributes(
        self, pydicom_attr, force_float=False, force_int=False
    ):
        """Generic method to get the values of a pydicom attribute for
        ControlPointSequence. This is used anytime you need to retrieve values
        that are specified for every control point

        Parameters
        ----------
        pydicom_attr : str
            the attribute of interest
        force_float : bool, optional
            force values to be of type float
        force_int : bool, optional
            force values to eb of type int

        Returns
        -------
        list
            the values of the specified attribute across all control points

        """
        values = []
        for cp in self.cp_seq:
            if hasattr(cp, pydicom_attr):
                if "Rotation" in pydicom_attr:
                    if getattr(cp, pydicom_attr).upper() in {"CC", "CW"}:
                        values.append(getattr(cp, pydicom_attr).upper())
                else:
                    value = getattr(cp, pydicom_attr)
                    if force_float:
                        value = float(value)
                    if force_int:
                        value = int(float(value))
                    values.append(value)

        if pydicom_attr[-5:] == "Angle":
            values = change_angle_origin(
                values, 180
            )  # if angle is greater than 180, convert to negative value
        return values

    def get_angle_values(self, angle_type):
        """Collect data to be used for database import related to objects that
        have angles

        Parameters
        ----------
        angle_type : str
            gantry', 'collimator' or 'couch'

        Returns
        -------
        dict
            start, end, rot_dir, range, min, and max values

        """
        angles = getattr(self, "%s_angles" % angle_type)
        return {
            "start": angles[0],
            "end": angles[-1],
            "rot_dir": self.get_rotation_direction(
                getattr(self, "%s_rot_dirs" % angle_type)
            ),
            "range": round(float(np.sum(np.abs(np.diff(angles)))), 1),
            "min": min(angles),
            "max": max(angles),
        }

    @staticmethod
    def get_rotation_direction(rotation_list):
        """Given a list of rotation directions, return either the only
        direction, or both with the initial direction first

        Parameters
        ----------
        rotation_list : list
            rotation angles of a beam attribute

        Returns
        -------
        str
            concise description of rotation

        """
        if not rotation_list:
            return None
        if len(set(rotation_list)) == 1:  # Only one direction found
            return rotation_list[0]
        return ["CC/CW", "CW/CC"][rotation_list[0] == "CW"]

    ###########################################################################
    # Properties
    ###########################################################################
    @property
    def beam_number(self):
        """Get the beam number

        Returns
        -------
        int
            BeamNumber (300A,00C0) converted to an int
        """
        return int(self.beam_data.BeamNumber)

    @property
    def beam_name(self):
        """Get the beam name

        Returns
        -------
        str
            Returns BeamDescription (300A,00C3), BeamName (300A,00C2), or
            BeamNumber (300A,00C0), in order of priority
        """
        if hasattr(self.beam_data, "BeamDescription"):
            return self.beam_data.BeamDescription
        if hasattr(self.beam_data, "BeamName"):
            return self.beam_data.BeamName
        return self.beam_number

    @property
    def treatment_machine(self):
        """Get the treatment for this beam

        Returns
        -------
        str
            TreatmentMachineName (300A,00B2)
        """
        return self.get_data_attribute(self.beam_data, "TreatmentMachineName")

    @property
    def beam_dose(self):
        """Get the beam dose

        Returns
        -------
        float
            BeamDose (300A,0084)
        """
        return self.get_data_attribute(
            self.ref_beam_data, "BeamDose", default=0.0, data_type="float"
        )

    @property
    def beam_mu(self):
        """Get the beam monitor units

        Returns
        -------
        float
            BeamMeterset (300A,0086)
        """
        return self.get_data_attribute(
            self.ref_beam_data, "BeamMeterset", default=0.0, data_type="float"
        )

    @property
    def beam_dose_pt(self):
        """Get the beam dose specification point

        Returns
        -------
        str
            Value from BeamDoseSpecificationPoint (300A,0082).
            NOTE: DICOM has since retired this tag.
        """
        return self.get_point_attribute(
            self.ref_beam_data, "BeamDoseSpecificationPoint"
        )

    @property
    def isocenter(self):
        """Get the isocenter coordintes

        Returns
        -------
        str
            csv of the isocenter coordinates in IsocenterPosition (300A,012C),
            for the first control point
        """
        return self.get_point_attribute(self.cp_seq[0], "IsocenterPosition")

    @property
    def isocenter_np(self):
        """Get the isocenter position as a numpy array

        Returns
        -------
        np.ndarray
            IsocenterPosition (300A,012C) for the first control point
        """
        return np.array(
            self.get_data_attribute(self.cp_seq[0], "IsocenterPosition")
        )

    @property
    def beam_type(self):
        """Get the beam type

        Returns
        -------
        str
            BeamType (300A,00C4)
        """
        return self.get_data_attribute(self.beam_data, "BeamType")

    @property
    def radiation_type(self):
        """Get the radiation type

        Returns
        -------
        str
            RadiationType (300A,00C6)
        """
        return self.get_data_attribute(self.beam_data, "RadiationType")

    @property
    def scan_mode(self):
        """Returns the scan mode (proton only)

        Returns
        -------
        str
            ScanMode (300A,0308)
        """
        return self.get_data_attribute(self.beam_data, "ScanMode")

    @property
    def scan_spot_count(self):
        """Get the number of scan spot positions (proton only)

        Returns
        -------
        int
            Sum of NumberOfScanSpotPositions (300A,0392)
        """
        if hasattr(self.cp_seq[0], "NumberOfScanSpotPositions"):
            return (
                sum([int(cp.NumberOfScanSpotPositions) for cp in self.cp_seq])
                / 2
            )

    @property
    def energies(self):
        """Collect the nominal beam energies across control points

        Returns
        -------
        list
            NominalBeamEnergy (300A,0114) for each control point
        """
        return self.get_cp_attributes("NominalBeamEnergy")

    @property
    def energy_min(self):
        """Get the minimum energy control point energy

        Returns
        -------
        float
            Minimum value in BeamParser.energies, rounded to 2 decimals
        """
        return round(min(self.energies), 2)

    @property
    def energy_max(self):
        """Get the maximum energy control point energy

        Returns
        -------
        float
            Maximum value in BeamParser.energies, rounded to 2 decimals
        """
        return round(max(self.energies), 2)

    @property
    def gantry_angles(self):
        """Get all gantry angles

        Returns
        -------
        list
            GantryAngle (300A,011E) for each control point
        """
        return self.get_cp_attributes("GantryAngle", force_float=True)

    @property
    def collimator_angles(self):
        """Get all collimator angles

        Returns
        -------
        list
            BeamLimitingDeviceAngle (300A,0120) for each control point
        """
        return self.get_cp_attributes(
            "BeamLimitingDeviceAngle", force_float=True
        )

    @property
    def couch_angles(self):
        """Get all couch angles

        Returns
        -------
        list
            PatientSupportAngle (300A,0122) for each control point
        """
        return self.get_cp_attributes("PatientSupportAngle", force_float=True)

    @property
    def gantry_rot_dirs(self):
        """Get all gantry rotation angles

        Returns
        -------
        list
            GantryRotationDirection (300A,011F) for each control point
        """
        return self.get_cp_attributes(
            "GantryRotationDirection", force_float=True
        )

    @property
    def collimator_rot_dirs(self):
        """Get all collimator rotation angles

        Returns
        -------
        list
            BeamLimitingDeviceRotationDirection (300A,0121) for each control
            point
        """
        return self.get_cp_attributes("BeamLimitingDeviceRotationDirection")

    @property
    def couch_rot_dirs(self):
        """Get all couch rotation angles

        Returns
        -------
        list
            PatientSupportRotationDirection (300A,0123) for each control point
        """
        return self.get_cp_attributes("PatientSupportRotationDirection")

    @property
    def gantry_values(self):
        """Get all gantry values

        Returns
        -------
        dict
            A collection of gantry values with keys of 'start', 'end',
            'rot_dir', 'range', 'min' and 'max'
        """
        return self.get_angle_values("gantry")

    @property
    def collimator_values(self):
        """Get all collimator values

        Returns
        -------
        dict
            A collection of collimator values with keys of 'start', 'end',
            'rot_dir', 'range', 'min' and 'max'
        """
        return self.get_angle_values("collimator")

    @property
    def couch_values(self):
        """Get all couch values

        Returns
        -------
        dict
            A collection of couch values with keys of 'start', 'end',
            'rot_dir', 'range', 'min' and 'max'
        """
        return self.get_angle_values("couch")

    @property
    def is_arc(self):
        """Get whether beam is an arc or not

        Returns
        -------
        bool
            True if beam contains rotation directions
        """

        return bool(self.gantry_values["rot_dir"])

    @property
    def tx_modality(self):
        """Get the tretament modality

        Returns
        -------
        str
            Returns BeamParser.radiation_type, modified with 3D or arc if
            BeamParser.is_arc is true
        """
        rad_type = str(self.radiation_type)
        if not rad_type:
            return None
        if "brachy" in rad_type.lower():
            return rad_type
        return "%s %s" % (
            self.radiation_type.title(),
            ["3D", "Arc"][self.is_arc],
        )

    @property
    def control_point_count(self):
        """Get the number of control points

        Returns
        -------
        int
            NumberOfControlPoints (300A,0110)
        """
        return int(
            self.get_data_attribute(self.beam_data, "NumberOfControlPoints")
        )

    @property
    def ssd(self):
        """Get the SSD of the beam

        Returns
        -------
        float
            SourceToSurfaceDistance (300A,0130), if beam is an arc then this
            will return the average over all control points
        """
        ssds = self.get_cp_attributes(
            "SourceToSurfaceDistance", force_float=True
        )
        if ssds:
            return round(float(np.average(ssds)) / 10.0, 2)

    @property
    def beam_mu_per_deg(self):
        """Beam monitor units divided by range of rotation (deg)

        Returns
        -------
        float
            BeamParser.beam_mu / BeamParser.gantry_values['range'] rounded
            to 2 decimals
        """
        try:
            return round(self.beam_mu / self.gantry_values["range"], 2)
        except Exception:
            pass

    @property
    def beam_mu_per_cp(self):
        """Beam monitor units divided by number of control points

        Returns
        -------
        float
            BeamParser.beam_mu / BeamParser.control_point_count rounded
            to 2 decimals
        """
        try:
            return round(self.beam_mu / self.control_point_count, 2)
        except Exception:
            pass

    @property
    def mlc_stat_data(self):
        """A collection of MLC metrics

        Returns
        -------
        dict
            MLC metrics with keys of 'area', 'x_perim', 'y_perim', 'perim',
            'cmp_score', 'cp_mu' based on dvha.tools.mlc_analyzer
        """
        mlc_keys = [
            "area",
            "x_perim",
            "y_perim",
            "perim",
            "cmp_score",
            "cp_mu",
        ]
        try:
            mlc_summary_data = mlca(
                self.beam_data,
                self.beam_mu,
                ignore_zero_mu_cp=True,
                **self.options
            ).summary
            mlca_stat_data = {
                key: calc_stats(mlc_summary_data[key]) for key in mlc_keys
            }
            mlca_stat_data["complexity"] = np.sum(
                mlc_summary_data["cmp_score"]
            )
        except Exception as e:
            push_to_log(
                e, msg="BeamParser: skipping mlc_stat_data calculation"
            )
            mlca_stat_data = {key: [None] * 6 for key in mlc_keys}
            mlca_stat_data["complexity"] = None
        return mlca_stat_data


class RxParser:
    """This class is used to parse rx data needed for importing a plan into
    the database

    Parameters
    ----------
    rt_plan : pydicom.DataSet
        the pydicom dataset object from reading the RT Plan DICOM file
    dicompyler_plan : dicompylercore.dicomparser.DicomParser
        the dicompyler-core object from reading the RT Plan DICOM file
    rt_structure : pydicom.DataSet
        the pydicom dataset object from reading the RT Structure DICOM file
    fx_grp_index : int
        the index of the fraction group associated with the rx to be parsed
    pinnacle_rx_data : dict
        prescription data extracted from dummy POIs
    study_instance_uid : str
        The study instance UID of the associated plan

    """

    def __init__(
        self,
        rt_plan,
        dicompyler_plan,
        rt_structure,
        fx_grp_index,
        pinnacle_rx_data,
        study_instance_uid,
    ):

        self.rt_plan = rt_plan
        self.dicompyler_plan = dicompyler_plan
        self.rt_structure = rt_structure
        self.study_instance_uid = study_instance_uid
        self.dose_ref_index = self.get_dose_ref_seq_index()

        if hasattr(rt_plan, "FractionGroupSequence"):
            self.fx_grp_data = rt_plan.FractionGroupSequence[fx_grp_index]
        else:
            self.fx_grp_data = None

        self.pinnacle_rx_data = None
        if pinnacle_rx_data and fx_grp_index + 1 in list(pinnacle_rx_data):
            self.pinnacle_rx_data = pinnacle_rx_data[fx_grp_index + 1]

    ###########################################################################
    # General rx parser tools
    ###########################################################################
    def get_dose_ref_seq_index(self):
        """ """
        if self.has_dose_ref:
            for i, dose_ref in enumerate(self.rt_plan.DoseReferenceSequence):
                if dose_ref.DoseReferenceNumber == self.fx_grp_number:
                    return i

    def get_dose_ref_attr(self, pydicom_attr):
        """Get a value from DoseReferenceSequence

        Parameters
        ----------
        pydicom_attr : str
            Any attribute in DoseReferenceSequence

        Returns
        -------
        any
            Value of ``pydicom_attr`` from DoseReferenceSequence (300A,0010)
            based on RxParser.dose_ref_index
        """
        if self.has_dose_ref and self.dose_ref_index is not None:
            dose_ref_data = self.rt_plan.DoseReferenceSequence[
                self.dose_ref_index
            ]
            if hasattr(dose_ref_data, pydicom_attr):
                return getattr(dose_ref_data, pydicom_attr)

    ###########################################################################
    # Properties
    ###########################################################################
    @property
    def plan_name(self):
        """Get the plan name

        Returns
        -------
        str
            RTPlanLabel (300A,0002)
        """
        if hasattr(self.rt_plan, "RTPlanLabel"):
            return self.rt_plan.RTPlanLabel

    @property
    def fx_grp_number(self):
        """If there are multiple RT Plan files for a given study instance uid,
        this function will ensure fx_grp_number is unique

        Returns
        -------
        int
            FractionGroupNumber (300A,0071) plus the max ``fx_grp_number``
            stored in the SQL database for this study instance uid

        """
        with DVH_SQL() as cnx:
            fraction_group_start = cnx.get_max_value(
                "Rxs",
                "fx_grp_number",
                "study_instance_uid = '%s'" % self.study_instance_uid,
            )
        if not fraction_group_start:
            fraction_group_start = 0

        if self.fx_grp_data is None:
            return fraction_group_start

        return self.fx_grp_data.FractionGroupNumber + fraction_group_start

    @property
    def fx_grp_name(self):
        """Get the fraction group name

        Returns
        -------
        str
            RxParser.fx_grp_number prepended by 'FxGrp '
        """
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data["fx_grp_name"]
        return "FxGrp %s" % self.fx_grp_number

    @property
    def has_dose_ref(self):
        """Check if DICOM-RT Plan as a DoseReferenceSequence

        Returns
        -------
        bool
            True if DICOM-RT Plan has DoseReferenceSequence (300A,0010)
        """
        return hasattr(self.rt_plan, "DoseReferenceSequence")

    @property
    def rx_dose(self):
        """Get the prescription dose

        Returns
        -------
        float
            In order of priority, get the prescription dose from
            RxParser.pinnacle_rx_data, TargetPrescriptionDose (300A,0026),
            prescription dose reported by dicompyler-core
        """
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data["rx_dose"]
        ans = self.get_dose_ref_attr("TargetPrescriptionDose")
        if ans is None:
            ans = float(self.dicompyler_plan["rxdose"]) / 100.0
        return ans

    @property
    def rx_percentage(self):
        """Get the prescription (Pinnacle only)

        Returns
        -------
        float
            Get RxParser.pinnacle_rx_data['rx_percentage'] if available
        """
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data["rx_percentage"]

    @property
    def fx_count(self):
        """Get the number of planned fractions

        Returns
        -------
        int
            NumberOfFractionsPlanned (300A,0078)
        """
        return int(
            float(getattr(self.fx_grp_data, "NumberOfFractionsPlanned", 0))
        )

    @property
    def fx_dose(self):
        """Get the fraction dose

        Returns
        -------
        float
            RxParser.pinnacle_rx_data['fx_dose'] if available, otherwise
            RxParser.rx_dose / RxParser.fx_count rounded to 2 decimals
        """
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data["fx_dose"]
        try:
            return round(self.rx_dose / float(self.fx_count), 2)
        except Exception as e:
            push_to_log(e, msg="RxParser: Unable to calculate fx_dose")

    @property
    def normalization_method(self):
        """Get the normalization method

        Returns
        -------
        str
            RxParser.pinnacle_rx_data['normalization_method'] if available,
            otherwise DoseReferenceStructureType (300A,0014)
        """
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data["normalization_method"]
        return self.get_dose_ref_attr("DoseReferenceStructureType")

    @property
    def normalization_object(self):
        """Get the normalization object

        Returns
        -------
        str
            In order of priority,
            RxParser.pinnacle_rx_data['normalization_object'],
            `COORDINATE` if normalization_method matches, or
            ManufacturerModelName
        """
        if self.pinnacle_rx_data:
            return self.pinnacle_rx_data["normalization_object"]
        if self.normalization_method:
            if self.normalization_method.lower() == "coordinates":
                return "COORDINATE"
            elif self.normalization_method.lower() == "site":
                if hasattr(self.rt_plan, "ManufacturerModelName"):
                    return self.rt_plan.ManufacturerModelName

    @property
    def beam_count(self):
        """Get the number of beams

        Returns
        -------
        int
            NumberOfBeams (300A,0080)
        """
        if hasattr(self.fx_grp_data, "NumberOfBeams"):
            return int(self.fx_grp_data.NumberOfBeams)
        return len(self.rt_plan.BeamSequence)

    @property
    def beam_numbers(self):
        """Get the beam numbers for the plan

        Returns
        -------
        list
            ReferencedBeamNumber (300C,0006)
        """
        if hasattr(self.fx_grp_data, "ReferencedBeamSequence"):
            return [
                ref_beam.ReferencedBeamNumber
                for ref_beam in self.fx_grp_data.ReferencedBeamSequence
            ]
        return list(range(1, self.beam_count + 1))


class PreImportData:
    """
    Light-weight object for the DICOM Importer GUI to avoid memory allocation
    issues
    """

    def __init__(
        self, file_set, stored_values, dicompyler_rt_structures, roi_map=None
    ):

        self.plan_file = file_set["plan"]
        self.dose_file = file_set["dose"]
        self.structure_file = file_set["structure"]
        self.missing_files = [
            key.capitalize()
            for key, value in file_set.items()
            if value is None
        ]

        self.stored_values = stored_values
        self.dicompyler_rt_structures = dicompyler_rt_structures
        self.database_rois = DatabaseROIs() if roi_map is None else roi_map

        # over ride objects
        keys = [
            "mrn",
            "study_instance_uid",
            "birth_date",
            "sim_study_date",
            "physician",
            "tx_site",
            "rx_dose",
        ]
        self.plan_over_rides = {key: None for key in keys}
        self.global_plan_over_rides = None

        self.roi_over_ride = {"name": {}, "type": {}}

        # If importing with auto sum turned off, use this list to track
        # plan-specific PTVs
        self.plan_ptvs = []

    @property
    def mrn(self):
        if self.plan_over_rides["mrn"] is not None:
            return self.plan_over_rides["mrn"]
        return self.stored_values["mrn"]

    @property
    def study_instance_uid(self):
        return self.stored_values["study_instance_uid_to_be_imported"]

    @property
    def study_instance_uid_to_be_imported(self):
        if (
            hasattr(self, "plan_over_rides")
            and self.plan_over_rides["study_instance_uid"]
        ):
            return self.plan_over_rides["study_instance_uid"]
        return self.stored_values["study_instance_uid_to_be_imported"]

    @property
    def is_study_instance_uid_valid(self):
        return self.is_uid_valid(self.study_instance_uid)

    @property
    def is_study_instance_uid_to_be_imported_valid(self):
        return self.is_uid_valid(self.study_instance_uid_to_be_imported)

    @staticmethod
    def is_uid_valid(uid):
        valid = False
        if uid:
            with DVH_SQL() as cnx:
                valid = not cnx.is_study_instance_uid_in_table("Plans", uid)
        return valid

    @property
    def patient_name(self):
        return self.stored_values["patient_name"]

    @property
    def file_set_complete(self):
        return all([self.plan_file, self.structure_file, self.dose_file])

    @property
    def rx_dose(self):
        if self.plan_over_rides["rx_dose"] is not None:
            ans = self.plan_over_rides["rx_dose"]
        else:
            ans = self.stored_values["rx_dose"]
        return self.process_global_over_ride("rx_dose", ans)

    @property
    def sim_study_date(self):
        if self.plan_over_rides["sim_study_date"] is not None:
            ans = self.plan_over_rides["sim_study_date"]
        else:
            ans = self.stored_values["sim_study_date"]
        return self.process_global_over_ride("sim_study_date", ans)

    @property
    def birth_date(self):
        if self.plan_over_rides["birth_date"] is not None:
            ans = self.plan_over_rides["birth_date"]
        else:
            ans = self.stored_values["birth_date"]
        return self.process_global_over_ride("birth_date", ans)

    @property
    def physician(self):
        if self.plan_over_rides["physician"] is not None:
            ans = self.plan_over_rides["physician"]
        else:
            ans = self.stored_values["physician"]
        return self.process_global_over_ride("physician", ans)

    @property
    def tx_site(self):
        if self.plan_over_rides["tx_site"] is not None:
            ans = self.plan_over_rides["tx_site"]
        else:
            ans = self.stored_values["tx_site"]
        return self.process_global_over_ride("tx_site", ans)

    @property
    def patient_orientation(self):
        return self.stored_values["patient_orientation"]

    def process_global_over_ride(self, key, pre_over_ride_value):
        if self.global_plan_over_rides:
            over_ride = self.global_plan_over_rides[key]
            if over_ride["value"]:
                if not over_ride["only_if_missing"] or (
                    over_ride["only_if_missing"] and not pre_over_ride_value
                ):
                    return over_ride["value"]
        return pre_over_ride_value

    def get_roi_type(self, key):
        """
        Get the ROI type as defined in DICOM (e.g., PTV, ORGAN, EXTERNAL, etc.)
        :param key: index of ROI
        :type key: int
        :return: the roi type
        :rtype: str
        """

        roi_type_from_roi_map = self.database_rois.get_roi_type(
            self.physician, self.get_physician_roi(key)
        )

        if key in list(self.roi_over_ride["type"]):
            ans = self.roi_over_ride["type"][key]
        elif roi_type_from_roi_map != "NONE":
            ans = roi_type_from_roi_map
        else:
            ans = self.dicompyler_rt_structures[key]["type"].upper()
        return ans if ans else "NONE"

    def reset_roi_type_over_ride(self, key):
        self.roi_over_ride["type"][key] = None

    def get_roi_name(self, key):
        """
        Applies clean_name from roi_name_manager.py to the name in the DICOM
        file
        :param key: the index of the roi
        :return: roi name to be used in the database
        :rtype: str
        """
        if key in list(self.roi_over_ride["name"]):
            return clean_name(self.roi_over_ride["name"][key])
        return clean_name(self.dicompyler_rt_structures[key]["name"])

    def set_roi_name(self, key, name):
        self.roi_over_ride["name"][key] = clean_name(name)

    def get_physician_roi(self, key):
        """
        Look up the physician roi of the specified roi for this plan
        :param key: the index of the roi
        :return: physician roi or 'uncategorized'
        :rtype: str
        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_physician_roi(
                    self.physician, roi_name
                )
        return "uncategorized"

    def get_institutional_roi(self, key):
        """
        Look up the institutional roi of the specified roi for this plan
        :param key: the index of the roi
        :return: institutional roi or 'uncategorized'
        :rtype: str
        """
        roi_name = self.get_roi_name(key)
        if self.database_rois.is_roi(roi_name):
            if self.database_rois.is_physician(self.physician):
                return self.database_rois.get_institutional_roi(
                    self.physician, self.get_physician_roi(key)
                )
            if roi_name in self.database_rois.institutional_rois:
                return roi_name
        return "uncategorized"

    @property
    def ptv_exists(self):
        """
        Check if the plan has at least one PTV assigned
        :rtype: bool
        """
        if self.dicompyler_rt_structures:
            for key in list(self.dicompyler_rt_structures):
                if self.get_roi_type(key) == "PTV":
                    return True
        return False

    @property
    def roi_names(self):
        """
        :return: roi_names for this plan per the plan (clean_name not applied)
        :rtype: list
        """
        return [
            self.get_roi_name(key)
            for key in list(self.dicompyler_rt_structures)
        ]

    def get_roi_key(self, roi_name):
        """
        Look-up the roi index for a given roi_name
        :param roi_name: the ROI name to be looked-up
        :return: the roi index or None
        :rtype: int
        """
        roi_name = clean_name(roi_name)
        for key in list(self.dicompyler_rt_structures):
            if roi_name == clean_name(self.get_roi_name(key)):
                return key

    @property
    def ptv_names(self):
        return self.stored_values["ptv_names"]

    @property
    def init_param(self):
        params = [
            "plan_file",
            "structure_file",
            "dose_file",
            "plan_ptvs",
            "plan_over_rides",
            "global_plan_over_rides",
            "roi_over_ride",
        ]
        return {key: getattr(self, key) for key in params}

    @property
    def validation(self):
        """
        :return: data used in GUI to indicate validity of data read-in from
        DICOM files to attract user to edit data
        :rtype: dict
        """
        return {
            "physician": {
                "status": self.database_rois.is_physician(self.physician)
                and self.physician != "DEFAULT",
                "value": self.physician,
                "message": "No physician assigned or physician is not in "
                "ROI Map.",
            },
            "mrn": {
                "status": self.mrn is not None,
                "value": self.mrn,
                "message": "MRN is empty.",
            },
            "study_instance_uid": {
                "status": self.is_study_instance_uid_to_be_imported_valid,
                "value": self.study_instance_uid_to_be_imported,
                "message": "Study Instance UID used in the database.",
            },
            "ptv": {
                "status": self.ptv_exists,
                "value": self.ptv_names,
                "message": "No PTV found.",
            },
            "sim_study_date": {
                "status": is_date(self.sim_study_date),
                "value": self.sim_study_date,
                "message": "Simulation date is empty or invalid.",
            },
            "rx_dose": {
                "status": self.rx_dose is not None,
                "value": self.rx_dose,
                "message": "Prescription dose is not defined.",
            },
            "complete_file_set": {
                "status": self.file_set_complete,
                "value": self.file_set_complete,
                "message": [
                    "Missing RT %s." % ", ".join(self.missing_files),
                    "",
                ][self.file_set_complete],
            },
        }

    @property
    def warning(self):
        msg = ""
        incomplete = False
        if self.plan_file and self.dose_file and self.structure_file:
            validation = self.validation
            failed_keys = {
                key for key, value in validation.items() if not value["status"]
            }
            if failed_keys:
                if "complete_file_set" in failed_keys:
                    msg = (
                        "ERROR: %s"
                        % validation["complete_file_set"]["message"]
                    )
                    incomplete = True
                else:
                    msg = "WARNING: %s" % " ".join(
                        [validation[key]["message"] for key in failed_keys]
                    )
        else:
            msg = (
                "ERROR: Incomplete Fileset. RT Plan, Dose, and Structure "
                "required."
            )
            incomplete = True

        return {"label": msg, "incomplete": incomplete}

    def autodetect_target_roi_type(self, key=None):
        """
        Target/tumor ROIs are often not labeled properly in ROI Type, but
        often start with correct acronym
        any ROI that starts with GTV, CTV, ITV, or PTV and is followed by
        nothing, a single number, or a character
        then a single number will be flagged.
        :param key: the index of the roi, if key is None, will search all keys
        """

        if key is None:
            roi_names = {
                key: structure["name"].lower()
                for key, structure in self.dicompyler_rt_structures.items()
            }
        else:
            roi_names = {
                key: self.dicompyler_rt_structures[key]["name"].lower()
            }

        targets = {"gtv", "ctv", "itv", "ptv"}
        for key, roi_name in roi_names.items():
            roi_name_len = len(roi_name)
            if (
                self.database_rois.get_roi_type(self.physician, roi_name)
                == "NONE"
            ):
                if self.get_physician_roi(key).lower() in targets:
                    self.roi_over_ride["type"][key] = self.get_physician_roi(
                        key
                    ).upper()
                elif (roi_name_len > 2 and roi_name[0:3] in targets) and (
                    (roi_name_len == 3)
                    or (roi_name_len == 4 and roi_name[3].isdigit())
                    or (
                        roi_name_len == 5
                        and not roi_name[3].isdigit()
                        and roi_name[4].isdigit()
                    )
                ):
                    self.roi_over_ride["type"][key] = roi_name[0:3].upper()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.dicom_dose_sum.py
"""
Class for summing dose grids
"""
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import numpy as np
from os.path import isfile
import pydicom
from scipy.ndimage import map_coordinates
from copy import deepcopy
from dvha.tools.errors import push_to_log


class DoseGrid:
    """
    Class to easily access commonly used attributes of a DICOM dose grid and
    perform summations

    Example: Add two dose grids
        grid_1 = DoseGrid(dose_file_1)
        grid_2 = DoseGrid(dose_file_2)
        grid_sum = grid_1 + grid_2
        grid_sum.save_dcm(some_file_path)

    """

    def __init__(
        self, rt_dose, order=1, try_full_interp=True, interp_block_size=50000
    ):
        """
        :param rt_dose: an RT Dose DICOM dataset or file_path
        :type rt_dose: pydicom.FileDataset
        :param order: The order of the spline interpolation. The order has to be in the range 0-5.
                      0: the nearest grid point, 1: trilinear, 2 to 5: spline
        :param try_full_interp: If true, will attempt to interpolate the entire grid at once before calculating one
                                block at a time (block size defined in self.interp_by_block)
        :type try_full_interp: bool
        :param interp_block_size: calculate this many points at a time if not try_full_interp or MemoryError
        :type interp_block_size: int
        """

        self.ds = self.__validate_input(rt_dose)
        self.order = order
        self.try_full_interp = try_full_interp
        self.interp_block_size = interp_block_size

        if self.ds:
            self.__set_axes()

    def __set_axes(self):
        self.x_axis = (
            np.arange(self.ds.Columns) * self.ds.PixelSpacing[0]
            + self.ds.ImagePositionPatient[0]
        )
        self.y_axis = (
            np.arange(self.ds.Rows) * self.ds.PixelSpacing[1]
            + self.ds.ImagePositionPatient[1]
        )
        self.z_axis = (
            np.array(self.ds.GridFrameOffsetVector)
            + self.ds.ImagePositionPatient[2]
        )

        # x and z are swapped in the pixel_array
        self.dose_grid = np.swapaxes(
            self.ds.pixel_array * self.ds.DoseGridScaling, 0, 2
        )

    @staticmethod
    def __validate_input(rt_dose):
        """Ensure provided input is either an RT Dose pydicom.FileDataset or a file_path to one"""
        if type(rt_dose) is pydicom.FileDataset:
            if rt_dose.Modality.lower() == "rtdose":
                return rt_dose
            msg = "DoseGrid.__validate_input: The provided pydicom.FileDataset is not RTDOSE"
            push_to_log(msg=msg)
            return
        elif isfile(rt_dose):
            try:
                rt_dose_ds = pydicom.read_file(rt_dose)
                if rt_dose_ds.Modality.lower() == "rtdose":
                    return rt_dose_ds
                msg = (
                    "DoseGrid.__validate_input: "
                    "The provided file_path points to a DICOM file, but it is not an RT Dose file."
                )
                push_to_log(msg=msg)
            except Exception as e:
                msg = (
                    "DoseGrid.__validate_input: "
                    "The provided input is neither a pydicom.FileDataset nor could it be read by pydicom."
                )
                push_to_log(e, msg=msg)
        return

    ####################################################
    # Basic properties
    ####################################################
    @property
    def shape(self):
        """Get the x, y, z dimensions of the dose grid"""
        return tuple(
            [self.ds.Columns, self.ds.Rows, len(self.ds.GridFrameOffsetVector)]
        )

    @property
    def axes(self):
        """Get the x, y, z axes of the dose grid (in mm)"""
        return [self.x_axis, self.y_axis, self.z_axis]

    @property
    def scale(self):
        """Get the dose grid resolution (xyz)"""
        return np.array(
            [
                self.ds.PixelSpacing[0],
                self.ds.PixelSpacing[1],
                self.ds.GridFrameOffsetVector[1]
                - self.ds.GridFrameOffsetVector[0],
            ]
        )

    @property
    def offset(self):
        """Get the coordinates of the dose grid origin (mm)"""
        return np.array(self.ds.ImagePositionPatient, dtype="float")

    @property
    def points(self):
        """Get all of the points in the dose grid"""
        y, x, z = np.meshgrid(self.y_axis, self.x_axis, self.z_axis)
        points = np.vstack((x.ravel(), y.ravel(), z.ravel()))
        return points.transpose()

    ####################################################
    # Tools
    ####################################################
    def __add__(self, other):
        """Addition in this fashion will not alter either DoseGrid, but it is more expensive with memory"""
        new = deepcopy(self)
        new.add(other)
        return new

    def is_coincident(self, other):
        """Check dose grid coincidence, if True a direct summation is appropriate"""
        return (
            self.ds.ImagePositionPatient == other.ds.ImagePositionPatient
            and self.ds.pixel_array.shape == other.ds.pixel_array.shape
            and self.ds.PixelSpacing == other.ds.PixelSpacing
            and self.ds.GridFrameOffsetVector == other.ds.GridFrameOffsetVector
        )

    def set_pixel_data(self):
        """
        Update the PixelData in the pydicom.FileDataset with the current self.dose_grid
        """
        self.ds.BitsAllocated = 32
        self.ds.BitsStored = 32
        self.ds.HighBit = 31
        self.ds.DoseGridScaling = (
            np.max(self.dose_grid) / np.iinfo(np.uint32).max
        )
        pixel_data = (
            np.swapaxes(self.dose_grid, 0, 2) / self.ds.DoseGridScaling
        )
        self.ds.PixelData = np.uint32(pixel_data).tostring()

    def save_dcm(self, file_path):
        """Save the pydicom.FileDataset to file"""
        self.ds.save_as(file_path)

    def get_ijk_points(self, other_axes):
        """
        Convert axes from another DoseGrid into ijk of this DoseGrid
        :param other_axes: the x, y, and z axis arrays
        :type other_axes: list
        :return: np.vstack of other_axes in this ijk space
        """
        ijk_axes = [
            (np.array(axis) - self.offset[a]) / self.scale[a]
            for a, axis in enumerate(other_axes)
        ]
        j, i, k = np.meshgrid(ijk_axes[1], ijk_axes[0], ijk_axes[2])
        return np.vstack((i.ravel(), j.ravel(), k.ravel()))

    ####################################################
    # Dose Summation
    ####################################################
    def add(self, other):
        """
        Add another 3D dose grid to this 3D dose grid, with interpolation if needed
        :param other: another DoseGrid
        :type other: DoseGrid
        """
        if self.is_coincident(other):
            self.direct_sum(other)
        else:
            self.interp_sum(other)

    def direct_sum(self, other, other_factor=1):
        """Directly sum two dose grids (only works if both are coincident)"""
        self.dose_grid += other.dose_grid * other_factor
        self.set_pixel_data()

    def interp_sum(self, other):
        """
        Interpolate the other dose grid to this dose grid's axes, then directly sum
        :param other: another DoseGrid
        :type other: DoseGrid
        """
        other_grid = None
        if self.try_full_interp:
            try:
                other_grid = self.interp_entire_grid(other)
            except MemoryError:
                pass
        if other_grid is None:
            other_grid = self.interp_by_block(other)

        self.dose_grid += other_grid
        self.set_pixel_data()

    def interp_entire_grid(self, other):
        """
        Interpolate the other dose grid to this dose grid's axes
        :param other: another DoseGrid
        :type other: DoseGrid
        """
        points = other.get_ijk_points(self.axes)
        return map_coordinates(
            input=other.dose_grid, coordinates=points, order=self.order
        ).reshape(self.shape)

    def interp_by_block(self, other):
        """
        Interpolate the other dose grid to this dose grid's axes, calculating one block at a time
        The block is defined at the init of this class, default is 50,000 points at a time
        :param other: another DoseGrid
        :type other: DoseGrid
        """
        points = other.get_ijk_points(self.axes)
        point_count = np.product(self.shape)
        other_grid = np.zeros(point_count)

        block_count = int(np.floor(point_count / self.interp_block_size))

        for i in range(block_count):
            start = i * self.interp_block_size
            end = (
                (i + 1) * self.interp_block_size if i + 1 < block_count else -1
            )
            other_grid[start:end] = map_coordinates(
                input=other.dose_grid,
                coordinates=points[start:end],
                order=self.order,
            )

        return other_grid.reshape(self.shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# options.py
"""
Class used to manage user options
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import pickle
from os.path import isfile
from os import unlink
import hashlib
from copy import deepcopy
from dvha._version import __version__
from dvha.paths import (
    OPTIONS_PATH,
    OPTIONS_CHECKSUM_PATH,
    INBOX_DIR,
    IMPORTED_DIR,
    REVIEW_DIR,
)
from dvha.tools.errors import push_to_log


class DefaultOptions:
    """Create default options, to be inherited by Options class"""

    def __init__(self):
        self.VERSION = __version__
        self.is_edited = False

        self.DB_TYPE = "sqlite"
        self.SQL_PGSQL_IP_HIST = []
        self.DEFAULT_CNF = {
            "pgsql": {"host": "localhost", "dbname": "dvh", "port": "5432"},
            "sqlite": {"host": "dvha.db"},
        }
        self.SQL_LAST_CNX = deepcopy(self.DEFAULT_CNF)

        self.DB_TYPE_GRPS = {grp: deepcopy(self.DB_TYPE) for grp in [1, 2]}
        self.SQL_LAST_CNX_GRPS = {
            1: deepcopy(self.DEFAULT_CNF),
            2: deepcopy(self.DEFAULT_CNF),
        }
        self.SYNC_SQL_CNX = True
        self._sql_vars = [
            "DB_TYPE",
            "SQL_PGSQL_IP_HIST",
            "DEFAULT_CNF",
            "SQL_LAST_CNX",
            "DB_TYPE_GRPS",
            "SQL_LAST_CNX_GRPS",
        ]

        self.MIN_BORDER = 50

        # These colors propagate to all tabs that visualize your two groups
        self.PLOT_COLOR = "blue"
        self.PLOT_COLOR_2 = "red"

        # The line width and style of selected DVHs in the DVH plot
        self.DVH_LINE_WIDTH_NONSELECTION = 2
        self.DVH_LINE_DASH_NONSELECTION = "solid"
        self.DVH_LINE_ALPHA_NONSELECTION = 0.075
        self.DVH_LINE_COLOR_NONSELECTION = "black"

        self.DVH_LINE_WIDTH_SELECTION = 4
        self.DVH_LINE_DASH_SELECTION = "solid"
        self.DVH_LINE_ALPHA_SELECTION = 1.0

        # Adjusts the opacity of the inner-quartile ranges
        self.IQR_ALPHA = 0.3

        # Adjust the plot font sizes
        self.PLOT_AXIS_LABEL_FONT_SIZE = "12pt"
        self.PLOT_AXIS_MAJOR_LABEL_FONT_SIZE = "10pt"

        # Grid line properties
        self.GRID_LINE_COLOR = "lightgrey"
        self.GRID_LINE_WIDTH = 1
        self.GRID_ALPHA = 1.0

        # Number of data points are reduced by this factor during dynamic
        # plot interaction to speed-up visualizations
        # This is only applied to the DVH plot since it has a large amount
        # of data
        self.LOD_FACTOR = 100

        # All DVHs in SQL DB have 1cGy bin widths regardless of this value.
        # However, the queried DVHs will be
        # down-sampled using this bin_width
        self.dvh_bin_width = 5

        # Passed into dicompyler-core to put a cap on the maximium dose to
        # prevent numpy.histogram from
        # blowing up memory allocation
        self.dvh_bin_max_dose = {"Gy": 500.0, "% Rx": 300.0}
        self.dvh_bin_max_dose_units = "Gy"
        self.dvh_bin_max_dose_options = ["Gy", "% Rx"]

        # Options for the group statistical DVHs in the DVHs tab
        self.STATS_MEDIAN_LINE_WIDTH = 2
        self.STATS_MEDIAN_LINE_DASH = "solid"
        self.STATS_MEDIAN_ALPHA = 0.6
        self.STATS_MEAN_LINE_WIDTH = 3
        self.STATS_MEAN_LINE_DASH = "dashed"
        self.STATS_MEAN_ALPHA = 0.5
        self.STATS_MAX_LINE_WIDTH = 2
        self.STATS_MAX_LINE_DASH = "dotted"
        self.STATS_MAX_ALPHA = 1
        self.STATS_MIN_LINE_WIDTH = 2
        self.STATS_MIN_LINE_DASH = "dotted"
        self.STATS_MIN_ALPHA = 1

        # Options for the time-series plot
        self.CORRELATION_POS_COLOR_1 = "blue"
        self.CORRELATION_NEG_COLOR_1 = "green"
        self.CORRELATION_POS_COLOR_2 = "red"
        self.CORRELATION_NEG_COLOR_2 = "purple"
        self.CORRELATION_MATRIX_VARS = [
            "Beam Area (Mean)",
            "Beam Dose (Mean)",
            "Beam MU (Mean)",
            "Beam Perimeter (Mean)",
            "PTV Cross-Section Median",
            "PTV Distance (Centroids)",
            "PTV Distance (Max)",
            "PTV Distance (Mean)",
            "PTV Distance (Median)",
            "PTV Distance (Min)",
            "PTV Max Dose",
            "PTV Min Dose",
            "PTV Surface Area",
            "PTV Volume",
            "Plan Complexity",
            "ROI Cross-Section Max",
            "ROI Cross-Section Median",
            "ROI Max Dose",
            "ROI Mean Dose",
            "ROI Min Dose",
            "ROI Surface Area",
            "ROI Volume",
            "Rx Dose",
            "Total Plan MU",
        ]

        # Options for the time-series plot
        self.TIME_SERIES_CIRCLE_SIZE = 10
        self.TIME_SERIES_CIRCLE_ALPHA = 0.3
        self.TIME_SERIES_TREND_LINE_WIDTH = 1
        self.TIME_SERIES_TREND_LINE_DASH = "solid"
        self.TIME_SERIES_AVG_LINE_WIDTH = 1
        self.TIME_SERIES_AVG_LINE_DASH = "dotted"
        self.TIME_SERIES_PATCH_ALPHA = 0.1

        # Options for the time-series plot
        self.CONTROL_CHART_CIRCLE_SIZE = 10
        self.CONTROL_CHART_CIRCLE_ALPHA = 0.3
        self.CONTROL_CHART_LINE_WIDTH = 1
        self.CONTROL_CHART_LINE_DASH = "solid"
        self.CONTROL_CHART_LINE_COLOR = "black"
        self.CONTROL_CHART_CENTER_LINE_WIDTH = 2
        self.CONTROL_CHART_CENTER_LINE_DASH = "solid"
        self.CONTROL_CHART_CENTER_LINE_COLOR = "black"
        self.CONTROL_CHART_CENTER_LINE_ALPHA = 1
        self.CONTROL_CHART_UCL_LINE_WIDTH = 2
        self.CONTROL_CHART_UCL_LINE_DASH = "dashed"
        self.CONTROL_CHART_UCL_LINE_COLOR = "red"
        self.CONTROL_CHART_UCL_LINE_ALPHA = 1
        self.CONTROL_CHART_LCL_LINE_WIDTH = 2
        self.CONTROL_CHART_LCL_LINE_DASH = "dashed"
        self.CONTROL_CHART_LCL_LINE_COLOR = "red"
        self.CONTROL_CHART_LCL_LINE_ALPHA = 1
        self.CONTROL_CHART_PATCH_ALPHA = 0.1
        self.CONTROL_CHART_PATCH_COLOR = "grey"
        self.CONTROL_CHART_OUT_OF_CONTROL_COLOR = "green"
        self.CONTROL_CHART_OUT_OF_CONTROL_COLOR_2 = "purple"
        self.CONTROL_CHART_OUT_OF_CONTROL_ALPHA = 1

        # Adjust the opacity of the histograms
        self.HISTOGRAM_ALPHA = 0.3

        # Options for the plot in the Multi-Variable Regression tab
        self.REGRESSION_CIRCLE_SIZE = 10
        self.REGRESSION_ALPHA = 0.5
        self.REGRESSION_LINE_WIDTH = 2
        self.REGRESSION_LINE_DASH = "dashed"

        self.REGRESSION_RESIDUAL_CIRCLE_SIZE = 3
        self.REGRESSION_RESIDUAL_ALPHA = 0.5
        self.REGRESSION_RESIDUAL_LINE_WIDTH = 2
        self.REGRESSION_RESIDUAL_LINE_DASH = "solid"
        self.REGRESSION_RESIDUAL_LINE_COLOR = "black"

        # Random forest
        self.MACHINE_LEARNING_ALPHA = 0.5
        self.MACHINE_LEARNING_ALPHA_DIFF = 0.35
        self.MACHINE_LEARNING_SIZE_PREDICT = 5
        self.MACHINE_LEARNING_SIZE_DATA = 5
        self.MACHINE_LEARNING_SIZE_MULTI_VAR = 5
        self.MACHINE_LEARNING_COLOR_PREDICT = "blue"
        self.MACHINE_LEARNING_COLOR_DATA = "black"
        self.MACHINE_LEARNING_COLOR_MULTI_VAR = "red"

        # This is the number of bins up do 100% used when resampling a DVH
        # to fractional dose
        self.RESAMPLED_DVH_BIN_COUNT = 5000

        self.MLC_ANALYZER_OPTIONS = {
            "max_field_size_x": 400.0,
            "max_field_size_y": 400.0,
            "complexity_weight_x": 1.0,
            "complexity_weight_y": 1.0,
        }

        # Per TG-263 (plus NONE, ITV, and IGNORED)
        self.ROI_TYPES = [
            "NONE",
            "ORGAN",
            "PTV",
            "ITV",
            "CTV",
            "GTV",
            "AVOIDANCE",
            "BOLUS",
            "CAVITY",
            "CONTRAST_AGENT",
            "EXTERNAL",
            "IRRAD_VOLUME",
            "REGISTRATION",
            "TREATED_VOLUME",
            "IGNORED",
        ]

        self.KEEP_IN_INBOX = 0
        self.SEARCH_SUBFOLDERS = 1
        self.IMPORT_UNCATEGORIZED = 0
        self.COPY_MISC_FILES = 0

        self.INBOX_DIR = INBOX_DIR
        self.IMPORTED_DIR = IMPORTED_DIR
        self.REVIEW_DIR = REVIEW_DIR

        self.MAX_DOSE_VOLUME = 0.03

        self.USE_DICOM_DVH = False
        self.AUTO_SUM_DOSE = True

        self.save_fig_param = {
            "figure": {
                "y_range_start": -0.0005,
                "x_range_start": 0.0,
                "y_range_end": 1.0005,
                "x_range_end": 10000.0,
                "background_fill_color": "none",
                "border_fill_color": "none",
                "plot_height": 600,
                "plot_width": 820,
            },
            "legend": {
                "background_fill_color": "white",
                "background_fill_alpha": 1.0,
                "border_line_color": "white",
                "border_line_alpha": 1.0,
                "border_line_width": 1,
            },
        }
        self.apply_range_edits = False

        self.positions = {
            "user_settings": None,
            "export_figure": None,
            "main": None,
        }
        self.window_sizes = {"main": None, "import": None}

        self.AUTO_SQL_DB_BACKUP = False

        self.MIN_RESOLUTION_MAIN = (1200, 700)
        self.MAX_INIT_RESOLUTION_MAIN = (1550, 900)

        self.SHOW_NEW_PTV_CALC_WARNING = True

        self.GET_DVH_KWARGS = {
            "calculate_full_volume": True,
            "use_structure_extents": False,
            "interpolation_resolution": None,
            "interpolation_segments_between_planes": 0,
            "memmap_rtdose": False,
        }
        self.DVH_SMALL_VOLUME_THRESHOLD = (
            10  # compute high resolution DVH if volume less than this (cc)
        )
        self.DVH_HIGH_RESOLUTION_FACTOR = 8  # Must be a factor of 2
        self.DVH_HIGH_RESOLUTION_FACTOR_OPTIONS = ["2", "4", "8", "16", "32"]
        self.DVH_HIGH_RESOLUTION_SEGMENTS_BETWEEN = 3  # Must be int


class Options(DefaultOptions):
    def __init__(self):
        DefaultOptions.__init__(self)
        self.__set_option_attr()

        self.load()

    def __set_option_attr(self):
        option_attr = []
        for attr in self.__dict__:
            if not attr.startswith("_"):
                option_attr.append(attr)
        self.option_attr = option_attr

    def load(self):
        self.is_edited = False
        if isfile(OPTIONS_PATH) and self.is_options_file_valid:
            try:
                with open(OPTIONS_PATH, "rb") as infile:
                    loaded_options = pickle.load(infile)
                self.upgrade_options(loaded_options)
            except Exception as e:
                msg = (
                    "Options.load: Options file corrupted. Loading "
                    "default options."
                )
                push_to_log(e, msg=msg)
                loaded_options = {}

            for key, value in loaded_options.items():
                if hasattr(self, key):
                    setattr(self, key, value)

    def save(self):
        self.is_edited = False
        out_options = {}
        for attr in self.option_attr:
            out_options[attr] = getattr(self, attr)
        out_options["VERSION"] = DefaultOptions().VERSION
        with open(OPTIONS_PATH, "wb") as outfile:
            pickle.dump(out_options, outfile)
        self.save_checksum()

    def set_option(self, attr, value):
        """
        Change or create an option value
        :param attr: name of option
        :type attr: str
        :param value: value of option
        """
        if not hasattr(self, attr):
            msg = "Options.set_option: %s did not previously exist" % attr
            push_to_log(msg=msg)
        setattr(self, attr, value)
        self.is_edited = True

    def save_checksum(self):
        check_sum = self.calculate_checksum()
        if check_sum:
            with open(OPTIONS_CHECKSUM_PATH, "w") as outfile:
                outfile.write(check_sum)

    @staticmethod
    def calculate_checksum():
        if isfile(OPTIONS_PATH):
            with open(OPTIONS_PATH, "rb") as infile:
                options_str = str(infile.read())
            return hashlib.md5(options_str.encode("utf-8")).hexdigest()
        return None

    @staticmethod
    def load_stored_checksum():
        if isfile(OPTIONS_CHECKSUM_PATH):
            with open(OPTIONS_CHECKSUM_PATH, "r") as infile:
                checksum = infile.read()
            return checksum
        return None

    @property
    def is_options_file_valid(self):
        try:
            current_checksum = self.calculate_checksum()
            stored_checksum = self.load_stored_checksum()
            if current_checksum == stored_checksum:
                return True
        except Exception as e:
            msg = (
                "Options.is_options_file_valid: Corrupted options file "
                "detected. Loading default options."
            )
            push_to_log(e, msg=msg)
            return False

    def restore_defaults(self):
        """Delete the store options file and checksum, load defaults"""
        if isfile(OPTIONS_PATH):
            unlink(OPTIONS_PATH)
        if isfile(OPTIONS_CHECKSUM_PATH):
            unlink(OPTIONS_CHECKSUM_PATH)
        default_options = DefaultOptions()

        for attr in default_options.__dict__:
            if not attr.startswith("_") and attr not in self._sql_vars:
                setattr(self, attr, getattr(default_options, attr))

    def clear_positions(self, *evt):
        """Clear all stored window positions, may be useful if window is
        off screen on Show"""
        self.positions = {key: None for key in list(self.positions)}

    def clear_window_sizes(self, *evt):
        """Clear all stored window sizes, may be useful if window is
        off screen on Show"""
        self.window_sizes = {key: None for key in list(self.window_sizes)}

    def apply_window_position(self, frame, position_key):
        """Given a frame, set to previously stored position or center it"""
        if self.positions[position_key] is not None:
            frame.SetPosition(self.positions[position_key])
        else:
            frame.Center()

    def set_window_size(self, frame, size_key):
        if size_key in self.window_sizes.keys():
            self.window_sizes[size_key] = frame.GetSize()

    def save_window_position(self, frame, position_key):
        """Store the position of the provided frame"""
        self.positions[position_key] = frame.GetPosition()

    def upgrade_options(self, loaded_options):
        """Reserve this space to apply all option file upgrades"""
        # This method is only needed for options that change type or structure
        # New options using a new attribute name will be automatically
        # generated by the DefaultOptions class
        self.db_group_upgrade(loaded_options)
        self.dvh_selection_upgrade(loaded_options)
        self.roi_type_upgrade(loaded_options)
        self.positions_upgrade(loaded_options)

    def positions_upgrade(self, loaded_options):
        if "main" not in loaded_options["positions"]:
            loaded_options["positions"]["main"] = None

    def db_group_upgrade(self, loaded_options):
        """DVHA v0.8.1 has a SQL cnx per group"""
        for key in ["DB_TYPE", "SQL_LAST_CNX"]:
            new_key = key + "_GRPS"
            if new_key not in loaded_options.keys():

                # DVHA <0.6.7 did not have DB_TYPE or SQL_LAST_CNX
                backup_value = (
                    "pgsql" if key == "DB_TYPE" else getattr(self, key)
                )  # sqlite not supported <0.6.7
                new_value = (
                    loaded_options[key]
                    if key in loaded_options.keys()
                    else backup_value
                )

                if sorted(list(new_value)) == [
                    1,
                    2,
                ]:  # users who may have used the dev branch
                    loaded_options[new_key] = {
                        grp: deepcopy(new_value[grp]) for grp in [1, 2]
                    }
                else:
                    loaded_options[new_key] = {
                        grp: deepcopy(new_value) for grp in [1, 2]
                    }

    def dvh_selection_upgrade(self, loaded_options):
        # Making nonselection dvh lines visible as v0.8.3
        # The following keys need updates to look nice, user can always
        # change later
        if "DVH_LINE_WIDTH_SELECTION" not in list(loaded_options):
            keys = [
                "IQR_ALPHA",
                "STATS_MEDIAN_LINE_WIDTH",
                "STATS_MEAN_LINE_WIDTH",
                "STATS_MAX_LINE_WIDTH",
                "STATS_MIN_LINE_WIDTH",
            ]
            for key in keys:
                loaded_options[key] = getattr(self, key)

    @staticmethod
    def roi_type_upgrade(loaded_options):
        """DVHA v0.8.3 added ROI Types into the ROI MAP, and added NONE and
        IGNORED types"""
        if "NONE" not in loaded_options["ROI_TYPES"]:
            loaded_options["ROI_TYPES"].insert(0, "NONE")
        if "IGNORED" not in loaded_options["ROI_TYPES"]:
            loaded_options["ROI_TYPES"].append("IGNORED")
